"""Collection and resolution API routes for the Matchbox server."""

import uuid
from typing import Annotated

from fastapi import (
    APIRouter,
    BackgroundTasks,
    Body,
    Depends,
    HTTPException,
    Query,
    UploadFile,
    status,
)
from pyarrow import ArrowInvalid
from pyarrow import parquet as pq

from matchbox.common.arrow import SCHEMA_INDEX, SCHEMA_RESULTS, table_to_buffer
from matchbox.common.dtos import (
    Collection,
    CollectionName,
    CRUDOperation,
    ModelResolutionName,
    NotFoundError,
    Resolution,
    ResolutionName,
    ResolutionPath,
    ResolutionType,
    ResourceOperationStatus,
    Run,
    RunID,
    UploadInfo,
    UploadStage,
)
from matchbox.common.exceptions import (
    MatchboxCollectionAlreadyExists,
    MatchboxCollectionNotFoundError,
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionAlreadyExists,
    MatchboxResolutionNotFoundError,
    MatchboxRunNotFoundError,
    MatchboxServerFileError,
)
from matchbox.server.api.dependencies import (
    BackendDependency,
    ParquetResponse,
    SettingsDependency,
    UploadTrackerDependency,
    authorisation_dependencies,
)
from matchbox.server.uploads import file_to_s3, process_upload, process_upload_celery

router = APIRouter(prefix="/collections", tags=["collection"])


# Collection management endpoints


@router.get(
    "",
    summary="List all collections",
    description="Retrieve a list of all collection names in the system.",
)
def list_collections(backend: BackendDependency) -> list[CollectionName]:
    """List all collections."""
    return backend.list_collections()


@router.get(
    "/{collection}",
    responses={404: {"model": NotFoundError}},
    summary="Get collection details",
    description=(
        "Retrieve details for a specific collection, including all its versions "
        "and resolutions."
    ),
)
def get_collection(
    backend: BackendDependency,
    collection: CollectionName,
) -> Collection:
    """Get collection details with all versions and resolutions."""
    return backend.get_collection(name=collection)


@router.post(
    "/{collection}",
    responses={
        409: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Create a new collection",
    description="Create a new collection with the specified name.",
)
def create_collection(
    backend: BackendDependency,
    collection: CollectionName,
) -> ResourceOperationStatus:
    """Create a new collection."""
    try:
        backend.create_collection(name=collection)
        return ResourceOperationStatus(
            success=True,
            target=f"Collection {collection}",
            operation=CRUDOperation.CREATE,
        )
    except MatchboxCollectionAlreadyExists as e:
        raise HTTPException(
            status_code=409,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Collection {collection}",
                operation=CRUDOperation.CREATE,
                details=str(e),
            ),
        ) from e


@router.delete(
    "/{collection}",
    responses={
        404: {"model": NotFoundError},
        409: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Delete a collection",
    description="Delete a collection and all its versions. Requires confirmation.",
)
def delete_collection(
    backend: BackendDependency,
    collection: CollectionName,
    certain: Annotated[
        bool, Query(description="Confirm deletion of the collection")
    ] = False,
) -> ResourceOperationStatus:
    """Delete a collection."""
    try:
        backend.delete_collection(name=collection, certain=certain)
        return ResourceOperationStatus(
            success=True,
            target=f"Collection {collection}",
            operation=CRUDOperation.DELETE,
        )
    except (
        MatchboxCollectionNotFoundError,
        MatchboxDeletionNotConfirmed,
    ):
        raise
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Collection {collection}",
                operation=CRUDOperation.DELETE,
                details=str(e),
            ),
        ) from e


# Run management endpoints


@router.get(
    "/{collection}/runs/{run_id}",
    responses={404: {"model": NotFoundError}},
    summary="Get specific run",
    description="Retrieve details for a specific run within a collection.",
)
def get_run(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
) -> Run:
    """Get a specific run."""
    return backend.get_run(collection=collection, run_id=run_id)


@router.post(
    "/{collection}/runs",
    responses={
        404: {"model": NotFoundError},
        409: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Create a new run",
    description="Create a new run within the specified collection.",
)
def create_run(
    backend: BackendDependency,
    collection: CollectionName,
) -> Run:
    """Create a new run in a collection."""
    try:
        return backend.create_run(collection=collection)
    except MatchboxCollectionNotFoundError as e:
        raise HTTPException(
            status_code=409,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Collection {collection}",
                operation=CRUDOperation.CREATE,
                details=str(e),
            ),
        ) from e


@router.patch(
    "/{collection}/runs/{run_id}/mutable",
    responses={
        404: {"model": NotFoundError},
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Change run mutability",
    description="Set whether a run can be modified.",
)
def set_run_mutable(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    mutable: Annotated[bool, Body(description="Mutability setting")],
) -> ResourceOperationStatus:
    """Set run mutability."""
    try:
        backend.set_run_mutable(collection=collection, run_id=run_id, mutable=mutable)
        return ResourceOperationStatus(
            success=True,
            target=f"Run {run_id}",
            operation=CRUDOperation.UPDATE,
        )
    except (
        MatchboxCollectionNotFoundError,
        MatchboxRunNotFoundError,
    ):
        raise
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Run {run_id}",
                operation=CRUDOperation.UPDATE,
                details=str(e),
            ).model_dump(),
        ) from e


@router.patch(
    "/{collection}/runs/{run_id}/default",
    responses={
        404: {"model": NotFoundError},
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Change default run",
    description="Set whether a run is the default for its collection.",
)
def set_run_default(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    default: Annotated[bool, Body(description="Default setting")],
) -> ResourceOperationStatus:
    """Set run as default."""
    try:
        backend.set_run_default(collection=collection, run_id=run_id, default=default)
        return ResourceOperationStatus(
            success=True,
            target=f"Run {run_id}",
            operation=CRUDOperation.UPDATE,
        )
    except (
        MatchboxCollectionNotFoundError,
        MatchboxRunNotFoundError,
    ):
        raise
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Run {run_id}",
                operation=CRUDOperation.UPDATE,
                details=str(e),
            ).model_dump(),
        ) from e


@router.delete(
    "/{collection}/runs/{run_id}",
    responses={
        404: {"model": NotFoundError},
        409: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Delete a run",
    description="Delete a run and all its resolutions. Requires confirmation.",
)
def delete_run(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    certain: Annotated[bool, Query(description="Confirm deletion of the run")] = False,
) -> ResourceOperationStatus:
    """Delete a run."""
    try:
        backend.delete_run(collection=collection, run_id=run_id, certain=certain)
        return ResourceOperationStatus(
            success=True, target=f"Run {run_id}", operation=CRUDOperation.DELETE
        )
    except (
        MatchboxCollectionNotFoundError,
        MatchboxRunNotFoundError,
        MatchboxDeletionNotConfirmed,
    ):
        raise
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Run {run_id}",
                operation=CRUDOperation.DELETE,
                details=str(e),
            ),
        ) from e


# Resolution management


@router.post(
    "/{collection}/runs/{run_id}/resolutions/{resolution_name}",
    responses={
        409: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Create a resolution",
    description="Create a new resolution (model or source) in the specified run.",
)
def create_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution_name: ResolutionName,
    resolution: Resolution,
) -> ResourceOperationStatus:
    """Create a resolution (model or source)."""
    resolution_path = ResolutionPath(
        name=resolution_name, collection=collection, run=run_id
    )
    try:
        backend.create_resolution(
            resolution=resolution,
            path=resolution_path,
        )
        return ResourceOperationStatus(
            success=True,
            target=f"Resolution {resolution_path}",
            operation=CRUDOperation.CREATE,
        )
    except (
        MatchboxCollectionNotFoundError,
        MatchboxRunNotFoundError,
    ):
        raise
    except MatchboxResolutionAlreadyExists as e:
        raise HTTPException(
            status_code=409,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution {resolution_path}",
                operation=CRUDOperation.CREATE,
                details=str(e),
            ),
        ) from e
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution {resolution_path}",
                operation=CRUDOperation.CREATE,
                details=str(e),
            ).model_dump(),
        ) from e


@router.get(
    "/{collection}/runs/{run_id}/resolutions/{resolution}",
    responses={404: {"model": NotFoundError}},
    summary="Get a resolution",
    description="Retrieve a specific resolution (model or source) from the backend.",
)
def get_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ResolutionName,
) -> Resolution:
    """Get a resolution (model or source) from the backend."""
    return backend.get_resolution(
        path=ResolutionPath(collection=collection, run=run_id, name=resolution)
    )


@router.put(
    "/{collection}/runs/{run_id}/resolutions/{resolution_name}",
    responses={
        404: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    status_code=status.HTTP_200_OK,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Update a resolution",
    description="Update an existing resolution (model or source) in the specified run.",
)
def update_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution_name: ResolutionName,
    resolution: Resolution,
) -> ResourceOperationStatus:
    """Update an existing resolution (model or source)."""
    resolution_path = ResolutionPath(
        name=resolution_name, collection=collection, run=run_id
    )
    try:
        backend.update_resolution(
            resolution=resolution,
            path=resolution_path,
        )
        return ResourceOperationStatus(
            success=True,
            target=f"Resolution {resolution_path}",
            operation=CRUDOperation.UPDATE,
        )
    except (
        MatchboxCollectionNotFoundError,
        MatchboxRunNotFoundError,
        MatchboxResolutionNotFoundError,
    ) as e:
        raise HTTPException(
            status_code=404,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution {resolution_path}",
                operation=CRUDOperation.UPDATE,
                details=str(e),
            ),
        ) from e
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution {resolution_path}",
                operation=CRUDOperation.UPDATE,
                details=str(e),
            ).model_dump(),
        ) from e


@router.delete(
    "/{collection}/runs/{run_id}/resolutions/{resolution}",
    responses={
        404: {"model": NotFoundError},
        409: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Delete a resolution",
    description="Delete a resolution from the backend. Requires confirmation.",
)
def delete_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ResolutionName,
    certain: Annotated[
        bool, Query(description="Confirm deletion of the resolution")
    ] = False,
) -> ResourceOperationStatus:
    """Delete a resolution from the backend."""
    resolution_path = ResolutionPath(collection=collection, run=run_id, name=resolution)
    try:
        backend.delete_resolution(path=resolution_path, certain=certain)
        return ResourceOperationStatus(
            success=True,
            target=f"Resolution {resolution_path}",
            operation=CRUDOperation.DELETE,
        )
    except (
        MatchboxCollectionNotFoundError,
        MatchboxRunNotFoundError,
        MatchboxDeletionNotConfirmed,
        MatchboxResolutionNotFoundError,
    ):
        raise
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution {resolution_path}",
                operation=CRUDOperation.DELETE,
                details=str(e),
            ),
        ) from e


@router.post(
    "/{collection}/runs/{run_id}/resolutions/{resolution_name}/data",
    responses={
        404: {"model": NotFoundError},
        400: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    status_code=status.HTTP_202_ACCEPTED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Set resolution data",
    description="Create an upload task for source hashes or model results.",
)
def set_data(
    backend: BackendDependency,
    upload_tracker: UploadTrackerDependency,
    settings: SettingsDependency,
    background_tasks: BackgroundTasks,
    collection: CollectionName,
    run_id: RunID,
    resolution_name: ModelResolutionName,
    file: UploadFile,
) -> ResourceOperationStatus:
    """Create an upload task for source hashes or model results."""
    resolution_path = ResolutionPath(
        collection=collection, run=run_id, name=resolution_name
    )
    # Not resistant to race conditions: currently, multiple requests to set data
    # could go through
    if backend.get_resolution_stage(path=resolution_path) != UploadStage.READY:
        raise HTTPException(
            status_code=400,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution data {resolution_path}",
                operation=CRUDOperation.CREATE,
                details="Not expecting upload for this resolution.",
            ),
        )
    resolution = backend.get_resolution(path=resolution_path)

    upload_id = str(uuid.uuid4())

    # Validate file
    if ".parquet" not in file.filename:
        raise HTTPException(
            status_code=400,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution data {resolution_path}",
                operation=CRUDOperation.CREATE,
                details=f"Expected .parquet file, got {file.filename.split('.')[-1]}",
            ),
        )

    # pyarrow validates Parquet magic numbers when loading file
    try:
        pq.ParquetFile(file.file)
    except ArrowInvalid as e:
        raise HTTPException(
            status_code=400,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution data {resolution_path}",
                operation=CRUDOperation.CREATE,
                details=f"Invalid Parquet file: {str(e)}",
            ),
        ) from e

    # Upload to S3
    client = backend.settings.datastore.get_client()
    bucket = backend.settings.datastore.cache_bucket_name
    key = f"{upload_id}.parquet"

    if resolution.resolution_type == ResolutionType.SOURCE:
        expected_schema = SCHEMA_INDEX
    else:
        expected_schema = SCHEMA_RESULTS

    table = pq.read_table(file.file)
    if not table.schema.equals(expected_schema):
        raise MatchboxServerFileError(
            message=(
                f"Schema mismatch. Expected:\n{expected_schema}\nGot:\n{table.schema}"
            )
        )
    try:
        file_to_s3(client=client, bucket=bucket, key=key, file=file)
    except MatchboxServerFileError as e:
        raise HTTPException(
            status_code=400,
            detail=ResourceOperationStatus(
                success=False,
                target=f"Resolution data {resolution_path}",
                operation=CRUDOperation.CREATE,
                details=f"Could not upload file to object storage: {str(e)}",
            ),
        ) from e

    # Start background processing
    backend.set_resolution_stage(path=resolution_path, stage=UploadStage.PROCESSING)
    match settings.task_runner:
        case "api":
            background_tasks.add_task(
                process_upload,
                backend=backend,
                tracker=upload_tracker,
                s3_client=client,
                resolution_path=resolution_path,
                upload_id=upload_id,
                bucket=bucket,
                filename=key,
            )
        case "celery":
            process_upload_celery.delay(
                resolution_path_json=resolution_path.model_dump_json(),
                upload_id=upload_id,
                bucket=bucket,
                filename=key,
            )
        case _:
            raise RuntimeError("Unsupported task runner.")

    return ResourceOperationStatus(
        success=True,
        target=f"Resolution data {resolution_path}",
        operation=CRUDOperation.CREATE,
        details=upload_id,
    )


@router.get(
    "/{collection}/runs/{run_id}/resolutions/{resolution}/data/status",
    responses={404: {"model": NotFoundError}},
    status_code=status.HTTP_200_OK,
)
def get_upload_status(
    backend: BackendDependency,
    upload_tracker: UploadTrackerDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ResolutionName,
    upload_id: Annotated[str | None, Query()] = None,
) -> UploadInfo:
    """Get the status of an upload process.

    Optionally looks for error message if given an upload ID.
    """
    error = None
    if upload_id:
        error = upload_tracker.get(upload_id=upload_id)

    resolution_stage = backend.get_resolution_stage(
        path=ResolutionPath(collection=collection, run=run_id, name=resolution)
    )

    return UploadInfo(stage=resolution_stage, error=error)


@router.get(
    "/{collection}/runs/{run_id}/resolutions/{resolution}/data",
    responses={404: {"model": NotFoundError}},
    summary="Get resolution results",
    description="Download results for a model as a parquet file.",
)
def get_results(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ModelResolutionName,
) -> ParquetResponse:
    """Download results for a model as a parquet file."""
    res = backend.get_model_data(
        path=ResolutionPath(collection=collection, run=run_id, name=resolution)
    )

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())



"""Base class for deduplication methodologies."""

import warnings
from abc import ABC, abstractmethod

import polars as pl
from pydantic import BaseModel, Field, ValidationInfo, field_validator


class DeduperSettings(BaseModel):
    """A data class to enforce basic settings dictionary shapes."""

    id: str = Field(default="id", description="A unique ID field in the data to dedupe")

    @field_validator("id")
    @classmethod
    def _id_for_mb(cls, v: str, info: ValidationInfo) -> str:
        enforce = "id"
        if v != enforce:
            warnings.warn(
                f"For offline deduplication, {info.field_name} can be any field. \n\n"
                "When deduplicating to write back to the Company Matching "
                f"Framework database, the ID must be '{enforce}', generated by "
                "matchbox queries.",
                stacklevel=3,
            )
        return v


class Deduper(BaseModel, ABC):
    """A base class for dedupers."""

    settings: DeduperSettings

    @abstractmethod
    def prepare(self, data: pl.DataFrame) -> None:
        """Prepare the deduper for deduplication."""
        return

    @abstractmethod
    def dedupe(self, data: pl.DataFrame) -> pl.DataFrame:
        """Deduplicate the dataframe."""
        return



"""Functions to compare fields in different sources."""

import sqlglot.expressions as exp
from sqlglot import parse_one
from sqlglot.errors import ParseError

from matchbox.common.logging import logger


def comparison(sql_condition: str, dialect: str = "postgres") -> str:
    """Validates any number of SQL conditions and prepares for a WHERE clause.

    Requires all column references be explicitly declared as from "l" and "r" tables.
    """
    parsed_sql = parse_one(sql_condition)

    for node in parsed_sql.walk():
        if not isinstance(
            node[0],
            exp.Connector | exp.Predicate | exp.Condition | exp.Identifier,
        ):
            raise ParseError(
                f"Must be valid WHERE clause statements. Found {type(node[0])}"
            )

        if isinstance(node[0], exp.Or):
            logger.warning(
                "Or conditions are computationally expensive. Expect slow performance."
            )

    left = False
    right = False
    for column in parsed_sql.find_all(exp.Column):
        if column.table == "l":
            left = True
        elif column.table == "r":
            right = True
        elif column.table == "":
            raise ParseError(
                "Columns must be explicitly declared as one of table "
                '"l" or "r".\n\n'
                f"No table declared for column {column.name}."
            )
        else:
            raise ParseError(
                "Columns must be explicitly declared as one of table "
                '"l" or "r".\n\n'
                f"Found column {column.table}."
            )
    if not (left and right):
        raise ParseError('Conditions must reference both "l" and "r".')

    pg_sql = parsed_sql.sql(dialect=dialect)

    return pg_sql



from unittest.mock import Mock, patch

import polars as pl
import pyarrow as pa
import pytest
from httpx import Response
from respx import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.locations import RelationalDBLocation
from matchbox.client.sources import (
    Source,
)
from matchbox.common.dtos import (
    BackendResourceType,
    CRUDOperation,
    DataTypes,
    NotFoundError,
    Resolution,
    ResourceOperationStatus,
    SourceField,
    UploadInfo,
    UploadStage,
)
from matchbox.common.exceptions import MatchboxServerFileError
from matchbox.common.factories.sources import (
    source_factory,
)


def test_source_infers_type(sqlite_warehouse: Engine) -> None:
    """Creating a source with type inference works."""
    # Create test data
    source_testkit = source_factory(
        n_true_entities=5,
        features=[
            {"name": "name", "base_generator": "word", "datatype": DataTypes.STRING},
        ],
        engine=sqlite_warehouse,
    ).write_to_location()

    location = RelationalDBLocation(name="dbname").set_client(sqlite_warehouse)
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=source_testkit.source_config.extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["name"],
    )

    assert source.config.key_field == SourceField(name="key", type=DataTypes.STRING)
    assert source.config.index_fields == tuple(
        [SourceField(name="name", type=DataTypes.STRING)]
    )


def test_source_sampling_preserves_original_sql(sqlite_warehouse: Engine) -> None:
    """SQL on RelationalDBLocation is preserved.

    SQLGlot transpiles INSTR() to STR_POSITION() in its default dialect.
    """
    # Create test data
    source_testkit = source_factory(
        n_true_entities=3,
        features=[
            {
                "name": "text_col",
                "base_generator": "word",
                "datatype": DataTypes.STRING,
            },
        ],
        engine=sqlite_warehouse,
    ).write_to_location()

    location = RelationalDBLocation(name="dbname").set_client(sqlite_warehouse)

    # Use SQLite's INSTR function (returns position of substring)
    # Other databases use CHARINDEX, POSITION, etc.
    extract_transform = f"""
        SELECT
            key,
            text_col,
            INSTR(text_col, 'a') as position_of_a
        FROM
            "{source_testkit.name}"
    """

    # This should work since INSTR is valid SQLite
    # Would fail if validation transpiles INSTR to POSITION() or similar
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["text_col", "position_of_a"],
    )

    assert source.config.key_field == SourceField(name="key", type=DataTypes.STRING)
    assert len(source.config.index_fields) == 2

    # This should work if the SQL is preserved exactly
    df = next(source.fetch())
    assert isinstance(df, pl.DataFrame)
    assert len(df) == 3


def test_source_fetch(sqlite_warehouse: Engine) -> None:
    """Test the query method with default parameters."""
    # Create test data
    source_testkit = source_factory(
        n_true_entities=5,
        features=[
            {"name": "name", "base_generator": "word", "datatype": DataTypes.STRING},
        ],
        engine=sqlite_warehouse,
    ).write_to_location()

    # Create location and source
    location = RelationalDBLocation(name="dbname").set_client(sqlite_warehouse)
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=source_testkit.source_config.extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["name"],
    )

    # Execute query
    result = next(source.fetch())

    # Verify result
    assert isinstance(result, pl.DataFrame)
    assert len(result) == 5
    assert "key" in result.columns
    assert "name" in result.columns

    # Try applying key filter
    key_subset = result[source.config.key_field.name][:2].to_list()
    result = next(source.fetch(keys=key_subset))
    assert len(result) == 2

    # Key filter ineffective with empty list
    result = next(source.fetch(keys=[]))
    assert len(result) == 5


@pytest.mark.parametrize(
    "qualify_names",
    [
        pytest.param(False, id="no_name_qualification"),
        pytest.param(True, id="with_name_qualification"),
    ],
)
@patch("matchbox.client.locations.RelationalDBLocation.execute")
def test_source_fetch_name_qualification(
    mock_execute: Mock, qualify_names: bool, sqlite_in_memory_warehouse: Engine
) -> None:
    """Test that column names are qualified when requested."""
    # Mock the location execute method to verify parameters
    mock_execute.return_value = (x for x in [None])  # execute needs to be a generator
    location = RelationalDBLocation(name="sqlite").set_client(
        sqlite_in_memory_warehouse
    )

    # Create source
    source = Source(
        dag=DAG("collection"),
        location=location,
        name="test_source",
        extract_transform="SELECT key, name FROM users",
        key_field=SourceField(name="key", type=DataTypes.STRING),
        index_fields=[SourceField(name="name", type=DataTypes.STRING)],
    )

    # Call query with qualification parameter
    next(source.fetch(qualify_names=qualify_names))

    # Verify the rename parameter passed to execute
    _, kwargs = mock_execute.call_args
    rename_param = kwargs.get("rename")

    if qualify_names:
        assert rename_param is not None
        assert callable(rename_param)
        # Test the rename function
        sample_col = "test_col"
        assert "test_source_" in source.name + "_" + sample_col
    else:
        assert rename_param is None


@pytest.mark.parametrize(
    ("batch_size", "expected_call_kwargs"),
    [
        pytest.param(
            None,
            {"batch_size": None},
            id="single_return",
        ),
        pytest.param(3, {"batch_size": 3}, id="multiple_batches"),
    ],
)
@patch("matchbox.client.locations.RelationalDBLocation.execute")
def test_source_fetch_batching(
    mock_execute: Mock,
    batch_size: int,
    expected_call_kwargs: dict,
    sqlite_in_memory_warehouse: Engine,
) -> None:
    """Test query with batching options."""
    # Mock the location execute method to verify parameters
    mock_execute.return_value = (x for x in [None])  # execute needs to be a generator
    location = RelationalDBLocation(name="sqlite").set_client(
        sqlite_in_memory_warehouse
    )

    # Create source
    source = Source(
        dag=DAG("collection"),
        location=location,
        name="test_source",
        extract_transform="SELECT key, name FROM users",
        key_field=SourceField(name="key", type=DataTypes.STRING),
        index_fields=[SourceField(name="name", type=DataTypes.STRING)],
    )

    # Call query with batching parameters
    next(source.fetch(batch_size=batch_size))

    # Verify parameters passed to execute
    _, kwargs = mock_execute.call_args
    for key, value in expected_call_kwargs.items():
        assert kwargs.get(key) == value


@pytest.mark.parametrize(
    "batch_size",
    [
        pytest.param(None, id="no_batching"),
        pytest.param(2, id="with_batching"),
    ],
)
def test_source_run(sqlite_warehouse: Engine, batch_size: int) -> None:
    """Test the run method produces expected hash format."""
    # Create test data with unique values
    n_true_entities = 3
    source_testkit = source_factory(
        n_true_entities=n_true_entities,
        features=[
            {"name": "name", "base_generator": "name", "datatype": DataTypes.STRING},
            {
                "name": "age",
                "base_generator": "random_int",
                "datatype": DataTypes.INT64,
            },
        ],
        engine=sqlite_warehouse,
    ).write_to_location()

    # Create location and source
    location = RelationalDBLocation(name="dbname").set_client(sqlite_warehouse)
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=source_testkit.source_config.extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["name", "age"],
    )

    # Execute run with different batching parameters
    result = source.run(batch_size=batch_size) if batch_size else source.run()

    # Verify result
    assert isinstance(result, pa.Table)
    assert "hash" in result.column_names
    assert "keys" in result.column_names
    assert len(result) == n_true_entities

    source.run()


@patch("matchbox.client.sources.Source.fetch")
def test_source_run_null_identifier(
    mock_fetch: Mock, sqlite_in_memory_warehouse: Engine
) -> None:
    """Test hashing data raises an error when source primary keys contain nulls."""
    # Create a source
    location = RelationalDBLocation(name="sqlite").set_client(
        sqlite_in_memory_warehouse
    )
    source = Source(
        dag=DAG("collection"),
        location=location,
        name="test_source",
        extract_transform="SELECT key, name FROM users",
        key_field=SourceField(name="key", type=DataTypes.STRING),
        index_fields=[SourceField(name="name", type=DataTypes.STRING)],
    )

    # Mock query to return data with null keys
    mock_df = pl.DataFrame({"key": ["1", None], "name": ["a", "b"]})
    mock_fetch.return_value = (x for x in [mock_df])

    # hashing data should raise ValueErrors for null keys
    with pytest.raises(ValueError, match="keys column contains null values"):
        source.run()


def test_source_sync(matchbox_api: MockRouter, sqlite_warehouse: Engine) -> None:
    """Test source syncing flow through the API."""
    # Mock source
    testkit = source_factory(
        features=[{"name": "company_name", "base_generator": "company"}],
        engine=sqlite_warehouse,
    ).write_to_location()

    # Mock the routes:
    # Resolution doesn't yet exist
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Source not found", entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        )
    )
    # Resolution can be inserted
    insert_config_route = matchbox_api.post(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            201,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.source.name}",
                operation=CRUDOperation.CREATE,
            ).model_dump_json(),
        )
    )
    # Resolution data can be inserted
    insert_hashes_route = matchbox_api.post(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data"
    ).mock(
        return_value=Response(
            202,
            json=ResourceOperationStatus(
                success=True, target="", operation=CRUDOperation.CREATE
            ).model_dump(),
        )
    )

    # Later, resolution can be updated
    update_route = matchbox_api.put(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            200,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.source.name}",
                operation=CRUDOperation.UPDATE,
            ).model_dump_json(),
        )
    )

    # Later, resolution can be deleted and recreated
    delete_route = matchbox_api.delete(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            200,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.source.name}",
                operation=CRUDOperation.DELETE,
            ).model_dump_json(),
        )
    )

    # -- ERRORS --

    # Can't sync before running
    with pytest.raises(RuntimeError, match="must be run"):
        testkit.source.sync()

    # We now run, but test that upload failure is handled
    testkit.fake_run()
    # Before and after upload, server accepts data - the second time reporting error
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(
        side_effect=[
            Response(200, json=UploadInfo(stage=UploadStage.READY).model_dump()),
            Response(
                200,
                json=UploadInfo(stage=UploadStage.READY, error="error").model_dump(),
            ),
        ]
    )
    with pytest.raises(MatchboxServerFileError, match="error"):
        testkit.source.sync()

    # -- FIRST TIME INSERTION --

    # Before upload, resolution is ready for data, after it is complete
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(
        side_effect=[
            Response(200, json=UploadInfo(stage=UploadStage.READY).model_dump()),
            Response(200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()),
        ]
    )

    # Sync the source, successfully
    testkit.source.run()
    testkit.source.sync()
    # Source was created, not updated or deleted
    assert insert_config_route.called
    assert not update_route.called
    assert not delete_route.called
    # Resolution metadata was correct
    resolution_call = Resolution.model_validate_json(
        insert_config_route.calls.last.request.content.decode("utf-8")
    )
    assert resolution_call == testkit.source.to_resolution()
    # Resolution data was correct
    assert (
        b"Content-Disposition: form-data;"
        in insert_hashes_route.calls.last.request.content
    )
    assert b"PAR1" in insert_hashes_route.calls.last.request.content

    # -- SOFT UPDATE --

    insert_hashes_route.reset()
    # Mock endpoint now returns existing resolution
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(return_value=Response(200, json=testkit.source.to_resolution().model_dump()))

    # Mock endpoint now declares data is present already
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(
        return_value=Response(
            200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()
        )
    )
    testkit.source.sync()
    # Resolution was compatible: ensure it was updated, not deleted
    assert update_route.called
    assert not delete_route.called
    # The data did not need to be updated
    assert not insert_hashes_route.called

    # -- HARD UPDATE --

    insert_hashes_route.reset()
    # Changing data requires deletion and re-insertion
    testkit.data_hashes = testkit.data_hashes.slice(1, 3)
    testkit.fake_run()
    # Resolution data is first ready to upload, and then uploaded
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(
        side_effect=[
            Response(200, json=UploadInfo(stage=UploadStage.READY).model_dump()),
            Response(200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()),
        ]
    )
    testkit.source.sync()

    assert delete_route.called
    assert insert_hashes_route.called



"""Unit tests for TestkitDAG."""

import pytest

from matchbox.common.factories.dags import TestkitDAG
from matchbox.common.factories.entities import FeatureConfig
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.sources import (
    SourceTestkitParameters,
    linked_sources_factory,
    source_factory,
)


def test_add_single_source() -> None:
    """Test adding a standalone source testkit."""
    dag_testkit = TestkitDAG()

    standalone = source_factory(
        name="test_source",
        features=[
            {"name": "company", "base_generator": "company"},
        ],
        dag=dag_testkit.dag,
    )

    dag_testkit.add_source(standalone)

    # Should store in sources registry
    assert "test_source" in dag_testkit.sources
    assert dag_testkit.sources["test_source"] == standalone

    # Should add to real DAG
    assert "test_source" in dag_testkit.dag.nodes
    assert dag_testkit.dag.nodes["test_source"] == standalone.source

    # Should not create linked entry
    assert len(dag_testkit.linked) == 0


def test_add_linked_sources() -> None:
    """Test adding a LinkedSourcesTestkit."""
    dag_testkit = TestkitDAG()

    features = [FeatureConfig(name="company", base_generator="company")]
    configs = (
        SourceTestkitParameters(name="source1", features=tuple(features)),
        SourceTestkitParameters(name="source2", features=tuple(features)),
    )

    linked = linked_sources_factory(
        source_parameters=configs, n_true_entities=5, dag=dag_testkit.dag
    )
    dag_testkit.add_linked_sources(linked)

    # Should store in linked registry with correct key
    expected_key = "linked_source1_source2"
    assert expected_key in dag_testkit.linked
    assert dag_testkit.linked[expected_key] == linked

    # Should store each source in sources registry
    assert "source1" in dag_testkit.sources
    assert "source2" in dag_testkit.sources
    assert dag_testkit.sources["source1"] == linked.sources["source1"]
    assert dag_testkit.sources["source2"] == linked.sources["source2"]

    # Should add both sources to real DAG
    assert "source1" in dag_testkit.dag.nodes
    assert "source2" in dag_testkit.dag.nodes


def test_add_model() -> None:
    """Test adding a model testkit."""
    dag_testkit = TestkitDAG()

    # Create source first
    linked = linked_sources_factory(n_true_entities=5, dag=dag_testkit.dag)
    dag_testkit.add_linked_sources(linked)

    # Create and add model
    model = model_factory(
        name="test_model",
        left_testkit=linked.sources["crn"],
        true_entities=tuple(linked.true_entities),
    )
    dag_testkit.add_model(model)

    # Should store in models registry
    assert "test_model" in dag_testkit.models
    assert dag_testkit.models["test_model"] == model

    # Should add to real DAG
    assert "test_model" in dag_testkit.dag.nodes
    assert dag_testkit.dag.nodes["test_model"] == model.model


def test_get_linked_testkit_for_source() -> None:
    """Test getting linked testkit for a source."""
    dag_testkit = TestkitDAG()

    # Add linked sources
    linked = linked_sources_factory(n_true_entities=5, dag=dag_testkit.dag)
    dag_testkit.add_linked_sources(linked)

    # Add standalone source
    standalone = source_factory(name="standalone", dag=dag_testkit.dag)
    dag_testkit.add_source(standalone)

    # Should find linked testkit for sources that belong to one
    linked_testkit = dag_testkit.source_to_linked["crn"]
    assert linked_testkit == linked

    linked_testkit = dag_testkit.source_to_linked["dh"]
    assert linked_testkit == linked

    # Standalone and nonexistent sources should generate a KeyError
    assert "standalone" not in dag_testkit.source_to_linked
    assert "nonexistent" not in dag_testkit.source_to_linked


def test_multiple_linked_testkits() -> None:
    """Test handling multiple LinkedSourcesTestkit objects."""
    dag_testkit = TestkitDAG()

    # Create two different linked testkits
    features = [
        FeatureConfig(name="company", base_generator="company"),
        FeatureConfig(name="email", base_generator="email"),
    ]

    configs1 = (
        SourceTestkitParameters(name="foo1", features=tuple(features[:1])),
        SourceTestkitParameters(name="foo2", features=tuple(features[:1])),
    )
    configs2 = (
        SourceTestkitParameters(name="bar1", features=tuple(features[1:])),
        SourceTestkitParameters(name="bar2", features=tuple(features[1:])),
    )

    linked1 = linked_sources_factory(
        source_parameters=configs1, n_true_entities=5, dag=dag_testkit.dag
    )
    linked2 = linked_sources_factory(
        source_parameters=configs2, n_true_entities=5, dag=dag_testkit.dag
    )

    dag_testkit.add_linked_sources(linked1)
    dag_testkit.add_linked_sources(linked2)

    # Expected linked keys
    assert "linked_foo1_foo2" in dag_testkit.linked
    assert "linked_bar1_bar2" in dag_testkit.linked
    assert dag_testkit.linked["linked_foo1_foo2"] == linked1
    assert dag_testkit.linked["linked_bar1_bar2"] == linked2

    # Sources should be registered separately
    assert len(dag_testkit.sources) == 4
    assert all(name in dag_testkit.sources for name in ["foo1", "foo2", "bar1", "bar2"])

    # Test linked testkit lookup works for each
    assert dag_testkit.source_to_linked["foo1"] == linked1
    assert dag_testkit.source_to_linked["bar1"] == linked2


def test_empty_dag_properties() -> None:
    """Test properties of an empty TestkitDAG."""
    dag_testkit = TestkitDAG()

    # All registries should be empty
    assert len(dag_testkit.sources) == 0
    assert len(dag_testkit.models) == 0
    assert len(dag_testkit.linked) == 0

    # get_linked_testkit should return None for anything
    assert "anything" not in dag_testkit.source_to_linked

    # DAG should be empty but valid
    assert len(dag_testkit.dag.nodes) == 0

    # final_step should raise on empty DAG
    with pytest.raises(ValueError):
        _ = dag_testkit.dag.final_step



from functools import partial

import polars as pl
import pyarrow as pa
import pyarrow.compute as pc
import pytest
from polars.testing import assert_frame_equal
from sqlalchemy import Engine

from matchbox.common.arrow import (
    SCHEMA_CLUSTER_EXPANSION,
    SCHEMA_EVAL_SAMPLES,
    SCHEMA_JUDGEMENTS,
    SCHEMA_QUERY,
    SCHEMA_QUERY_WITH_LEAVES,
)
from matchbox.common.dtos import (
    DataTypes,
    Match,
    ModelConfig,
    Resolution,
    ResolutionPath,
    SourceConfig,
    SourceField,
    UploadStage,
)
from matchbox.common.eval import Judgement
from matchbox.common.exceptions import (
    MatchboxCollectionAlreadyExists,
    MatchboxCollectionNotFoundError,
    MatchboxDataNotFound,
    MatchboxNoJudgements,
    MatchboxResolutionAlreadyExists,
    MatchboxResolutionExistingData,
    MatchboxResolutionNotFoundError,
    MatchboxResolutionUpdateError,
    MatchboxRunNotFoundError,
    MatchboxRunNotWriteable,
)
from matchbox.common.factories.entities import (
    SourceEntity,
    diff_results,
    query_to_cluster_entities,
)
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.scenarios import setup_scenario
from matchbox.common.factories.sources import SourceTestkit
from matchbox.server.base import MatchboxDBAdapter

backends = [
    pytest.param("matchbox_postgres", id="postgres"),
]


@pytest.fixture(scope="function")
def backend_instance(request: pytest.FixtureRequest, backend: str) -> MatchboxDBAdapter:
    """Create a fresh backend instance for each test."""
    backend_obj = request.getfixturevalue(backend)
    backend_obj.clear(certain=True)
    return backend_obj


@pytest.mark.parametrize("backend", backends)
@pytest.mark.docker
class TestMatchboxBackend:
    @pytest.fixture(autouse=True)
    def setup(self, backend_instance: str, sqlite_warehouse: Engine) -> None:
        self.backend: MatchboxDBAdapter = backend_instance
        self.scenario = partial(setup_scenario, warehouse=sqlite_warehouse)

    def test_properties(self) -> None:
        """Test that properties obey their protocol restrictions."""
        with self.scenario(self.backend, "index"):
            assert isinstance(self.backend.sources.list_all(), list)
            assert isinstance(self.backend.sources.count(), int)
            assert isinstance(self.backend.models.count(), int)
            assert isinstance(self.backend.data.count(), int)
            assert isinstance(self.backend.clusters.count(), int)
            assert isinstance(self.backend.creates.count(), int)
            assert isinstance(self.backend.merges.count(), int)
            assert isinstance(self.backend.proposes.count(), int)

    # Retrieval

    def test_query_only_source(self) -> None:
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")

            df_crn_sample = self.backend.query(
                source=crn_testkit.resolution_path,
                limit=10,
            )

            assert isinstance(df_crn_sample, pa.Table)
            assert df_crn_sample.num_rows == 10

            df_crn_full = self.backend.query(source=crn_testkit.resolution_path)

            assert df_crn_full.num_rows == crn_testkit.data.num_rows
            assert df_crn_full.schema.equals(SCHEMA_QUERY)

    def test_query_return_leaf_ids(self) -> None:
        """Test querying data and additionally requesting leaf IDs."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")

            df_crn_full = self.backend.query(
                source=crn_testkit.resolution_path, return_leaf_id=True
            )

            assert df_crn_full.num_rows == crn_testkit.data.num_rows
            assert df_crn_full.schema.equals(SCHEMA_QUERY_WITH_LEAVES)

    def test_query_with_dedupe_model(self) -> None:
        """Test querying data from a deduplication point of truth."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            df_crn = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.data.num_rows
            assert df_crn.schema.equals(SCHEMA_QUERY)

            linked = dag_testkit.source_to_linked["crn"]

            assert pc.count_distinct(df_crn["id"]).as_py() == len(
                linked.true_entity_subset("crn")
            )

    def test_query_with_link_model(self) -> None:
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            df_crn = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.data.num_rows
            assert df_crn.schema.equals(SCHEMA_QUERY)

            df_dh = self.backend.query(
                source=dh_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_dh, pa.Table)
            assert df_dh.num_rows == dh_testkit.data.num_rows
            assert df_dh.schema.equals(SCHEMA_QUERY)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            all_ids = pa.concat_arrays(
                [df_crn["id"].combine_chunks(), df_dh["id"].combine_chunks()]
            )

            assert pc.count_distinct(all_ids).as_py() == len(
                linked.true_entity_subset("crn", "dh")
            )

    def test_threshold_query_with_link_model(self) -> None:
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "probabilistic_naive_test_crn_naive_test_cdms"
            crn_testkit = dag_testkit.sources.get("crn")
            cdms_testkit = dag_testkit.sources.get("cdms")
            linker_testkit = dag_testkit.models.get(linker_name)

            df_crn = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.data.num_rows
            assert df_crn.schema.equals(SCHEMA_QUERY)

            df_cdms = self.backend.query(
                source=cdms_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_cdms, pa.Table)
            assert df_cdms.num_rows == cdms_testkit.data.num_rows
            assert df_cdms.schema.equals(SCHEMA_QUERY)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # Test query with threshold
            df_crn_threshold = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
                threshold=100,
            )
            df_cdms_threshold = self.backend.query(
                source=cdms_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
                threshold=100,
            )
            threshold_ids = pa.concat_arrays(
                [
                    df_crn_threshold["id"].combine_chunks(),
                    df_cdms_threshold["id"].combine_chunks(),
                ]
            )

            # Query returns more clusters when threshold exceeds
            # true entity match probabilities
            assert pc.count_distinct(threshold_ids).as_py() > len(
                linked.true_entity_subset("crn", "cdms")
            )

    def test_match_one_to_many(self) -> None:
        """Test that matching data works when the target has many IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random one:many entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "dh": 1},
                max_appearances={"dh": 1},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["dh"])),
                source=dh_testkit.resolution_path,
                targets=[crn_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == dh_testkit.source.resolution_path
            assert res[0].target == crn_testkit.source.resolution_path
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.keys["dh"]
            assert res[0].target_id == source_entity.keys["crn"]

    def test_match_many_to_one(self) -> None:
        """Test that matching data works when the source has more possible IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random many:one entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "dh": 1},
                max_appearances={"dh": 1},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["crn"])),
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.keys["crn"]
            assert res[0].target_id == source_entity.keys["dh"]

    def test_match_one_to_none(self) -> None:
        """Test that matching data works when the target has no IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random one:none entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 1},
                max_appearances={"dh": 0},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["crn"])),
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.keys["crn"]
            assert res[0].target_id == source_entity.keys.get("dh", set())

    def test_match_none_to_none(self) -> None:
        """Test that matching data works when the supplied key doesn't exist."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Use a non-existent source key
            non_existent_key = "foo"

            res = self.backend.match(
                key=non_existent_key,
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].cluster is None
            assert res[0].source_id == set()
            assert res[0].target_id == set()

    def test_threshold_match_many_to_one(self) -> None:
        """Test that matching data works when the target has many IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "probabilistic_naive_test_crn_naive_test_cdms"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random one:many entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "dh": 1},
                max_appearances={"dh": 1},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["crn"])),
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
                threshold=100,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].source_id == source_entity.keys["crn"]
            # Match does not return true target ids when threshold
            # exceeds match probability
            assert len(res[0].target_id) < len(source_entity.keys["dh"])

    # Collection management

    def test_collections(self) -> None:
        """Test creating, listing, getting and deleting collections."""
        with self.scenario(self.backend, "bare") as _:
            collections_pre = self.backend.list_collections()
            assert "test_collection" not in collections_pre

            # Create new collection and verify its initial properties
            test_collection_created = self.backend.create_collection("test_collection")

            assert test_collection_created.runs == []  # No versions yet

            collections_post = self.backend.list_collections()
            assert "test_collection" in collections_post

            # Verify duplicate creation is rejected
            with pytest.raises(MatchboxCollectionAlreadyExists):
                self.backend.create_collection("test_collection")

            test_collection = self.backend.get_collection("test_collection")

            assert test_collection == test_collection_created

            self.backend.delete_collection("test_collection", certain=True)

            # Verify collection is properly removed
            with pytest.raises(MatchboxCollectionNotFoundError):
                self.backend.get_collection("test_collection")

            with pytest.raises(MatchboxCollectionNotFoundError):
                self.backend.delete_collection("test_collection", certain=False)

    # Run management

    def test_runs(self) -> None:
        """Test creating, listing, getting and deleting runs."""
        with self.scenario(self.backend, "bare") as _:
            collections_pre = self.backend.list_collections()
            assert "test_collection" not in collections_pre

            # Create parent collection with no runs initially
            test_collection_pre = self.backend.create_collection("test_collection")
            assert len(test_collection_pre.runs) == 0  # No runs yet

            # Create first run and check default properties
            v1 = self.backend.create_run("test_collection")

            assert isinstance(v1.run_id, int)
            assert v1.is_default is False  # New runs aren't default by default
            assert v1.is_mutable is True  # New runs are mutable by default
            assert v1.resolutions == {}  # No resolutions yet

            test_collection_post = self.backend.get_collection("test_collection")
            assert v1.run_id in test_collection_post.runs

            v2 = self.backend.create_run("test_collection")

            with pytest.raises(ValueError, match="mutable"):
                self.backend.set_run_default("test_collection", v1.run_id, True)

            self.backend.set_run_mutable("test_collection", v1.run_id, False)
            self.backend.set_run_default("test_collection", v1.run_id, True)

            # Default run info also available on collection DTO
            collection = self.backend.get_collection("test_collection")
            assert collection.default_run == v1.run_id

            # Setting v2 as default should automatically unset v1 as default
            self.backend.set_run_mutable("test_collection", v2.run_id, False)
            self.backend.set_run_default("test_collection", v2.run_id, True)

            v1 = self.backend.get_run("test_collection", v1.run_id)
            v2 = self.backend.get_run("test_collection", v2.run_id)

            assert v1.is_mutable is False
            assert v1.is_default is False

            assert v2.is_mutable is False
            assert v2.is_default is True

            self.backend.delete_run("test_collection", v1.run_id, certain=True)

            # Verify run is properly removed
            with pytest.raises(MatchboxRunNotFoundError):
                self.backend.get_run("test_collection", v1.run_id)

            with pytest.raises(MatchboxRunNotFoundError):
                self.backend.delete_run("test_collection", v1.run_id, certain=False)

    def test_run_immutable(self) -> None:
        """Nothing in an immutable run can be changed."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            source_testkit = dag_testkit.sources["crn"]
            model_testkit = dag_testkit.models["naive_test_crn"]

            self.backend.set_run_mutable(
                collection=dag_testkit.dag.name,
                run_id=dag_testkit.dag.run,
                mutable=False,
            )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.create_resolution(
                    path=source_testkit.resolution_path.model_copy(
                        update={"name": "new_source"}
                    ),
                    resolution=source_testkit.fake_run().source.to_resolution(),
                )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.delete_resolution(
                    source_testkit.resolution_path, certain=True
                )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.insert_source_data(
                    source_testkit.resolution_path, source_testkit.data_hashes
                )

            with pytest.raises(MatchboxRunNotWriteable):
                self.backend.insert_model_data(
                    model_testkit.resolution_path,
                    model_testkit.probabilities.to_arrow(),
                )

    # Resolution management

    def test_get_source(self) -> None:
        """Test querying data from the database."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")

            crn_retrieved = self.backend.get_resolution(crn_testkit.resolution_path)
            assert isinstance(crn_retrieved, Resolution)
            assert crn_testkit.source_config == crn_retrieved.config

            with pytest.raises(MatchboxResolutionNotFoundError):
                self.backend.get_resolution(
                    path=ResolutionPath(collection="collection", run=1, name="foo")
                )

    def test_delete_resolution(self) -> None:
        """
        Tests the deletion of:

        * The resolution from the resolutions table
        * The source configuration attached to the resolution
        * Any models that depended on this model (descendants)
        * All probabilities for all descendants
        """
        with self.scenario(self.backend, "link") as dag_testkit:
            to_delete = dag_testkit.sources["crn"]
            total_sources = len(dag_testkit.sources)
            total_models = len(dag_testkit.models)

            source_configs_pre_delete = self.backend.sources.count()
            sources_pre_delete = self.backend.source_resolutions.count()
            models_pre_delete = self.backend.models.count()
            cluster_count_pre_delete = self.backend.clusters.count()
            cluster_assoc_count_pre_delete = self.backend.creates.count()
            proposed_merge_probs_pre_delete = self.backend.proposes.count()

            assert sources_pre_delete == total_sources
            assert models_pre_delete == total_models
            assert cluster_count_pre_delete > 0
            assert cluster_assoc_count_pre_delete > 0
            assert proposed_merge_probs_pre_delete > 0

            # Perform deletion
            self.backend.delete_resolution(to_delete.resolution_path, certain=True)

            source_configs_post_delete = self.backend.sources.count()
            sources_post_delete = self.backend.source_resolutions.count()
            models_post_delete = self.backend.models.count()
            cluster_count_post_delete = self.backend.clusters.count()
            cluster_assoc_count_post_delete = self.backend.creates.count()
            proposed_merge_probs_post_delete = self.backend.proposes.count()

            # 1 source, 1 index, (1 deduper + 3 linkers) = 4 models are gone
            assert source_configs_post_delete == source_configs_pre_delete - 1
            assert sources_post_delete == sources_pre_delete - 1
            assert models_post_delete == models_pre_delete - 4

            # We've lost some composite clusters
            assert cluster_count_post_delete < cluster_count_pre_delete

            # Count of propose and create edges has dropped
            assert cluster_assoc_count_post_delete < cluster_assoc_count_pre_delete
            assert proposed_merge_probs_post_delete < proposed_merge_probs_pre_delete

    def test_index_new_source(self) -> None:
        """Test that indexing a new source works."""
        with self.scenario(self.backend, "bare") as dag_testkit:
            crn_testkit: SourceTestkit = dag_testkit.sources.get("crn").fake_run()

            assert self.backend.clusters.count() == 0

            self.backend.create_resolution(
                crn_testkit.source.to_resolution(),
                path=crn_testkit.resolution_path,
            )

            # Resolution can't be re-added
            with pytest.raises(MatchboxResolutionAlreadyExists):
                self.backend.create_resolution(
                    crn_testkit.source.to_resolution(),
                    path=crn_testkit.resolution_path,
                )

            # After resolution metadata is present, we can add data
            self.backend.insert_source_data(
                crn_testkit.source.resolution_path, crn_testkit.data_hashes
            )

            # Data can't be re-added
            with pytest.raises(MatchboxResolutionExistingData):
                self.backend.insert_source_data(
                    crn_testkit.source.resolution_path, crn_testkit.data_hashes
                )

            # Resolution marked as complete
            assert (
                self.backend.get_resolution_stage(crn_testkit.source.resolution_path)
                == UploadStage.COMPLETE
            )

            # We can retrieve the resolution
            crn_retrieved = self.backend.get_resolution(
                crn_testkit.source.resolution_path
            )

            assert crn_testkit.source_config == crn_retrieved.config
            assert self.backend.data.count() == len(crn_testkit.data_hashes)

            assert self.backend.data.count() == len(crn_testkit.data_hashes)
            assert self.backend.source_resolutions.count() == 1

            # We can update the resolution metadata, including changes in fields
            updated_key_field = SourceField(name="new key", type=DataTypes.STRING)
            updated_index_fields = (
                SourceField(name="new field", type=DataTypes.BOOLEAN),
            )
            updated_config = SourceConfig.model_validate(
                crn_testkit.source.config.model_copy(
                    update={
                        "key_field": updated_key_field,
                        "index_fields": updated_index_fields,
                    }
                )
            )
            updated_resolution = Resolution.model_validate(
                crn_testkit.source.to_resolution().model_copy(
                    update={
                        "description": "updated",
                        "config": updated_config,
                    }
                )
            )
            self.backend.update_resolution(
                updated_resolution, path=crn_testkit.source.resolution_path
            )

            # We cannot update source resolution with different fingerprint
            with pytest.raises(MatchboxResolutionUpdateError, match="fingerprint"):
                self.backend.update_resolution(
                    updated_resolution.model_copy(update={"fingerprint": 123}),
                    path=crn_testkit.source.resolution_path,
                )

            # We cannot update source resolution with a model resolution
            with pytest.raises(MatchboxResolutionUpdateError, match="parents"):
                # Create model with same fingerprint as previous resolution
                valid_fingerprint = crn_testkit.source.to_resolution().fingerprint
                model_resolution = Resolution.model_validate(
                    model_factory()
                    .fake_run()
                    .model.to_resolution()
                    .model_copy(update={"fingerprint": valid_fingerprint})
                )
                self.backend.update_resolution(
                    model_resolution,
                    path=crn_testkit.source.resolution_path,
                )

            # We can retrieve the updated resolution
            crn_retrieved = self.backend.get_resolution(
                crn_testkit.source.resolution_path
            )
            assert crn_retrieved.description == "updated"
            assert crn_retrieved.config.key_field == updated_key_field
            assert crn_retrieved.config.index_fields == updated_index_fields

    def test_index_empty_source(self) -> None:
        """Can insert and retrieve empty source data"""
        with self.scenario(self.backend, "bare") as dag_testkit:
            crn_testkit: SourceTestkit = dag_testkit.sources.get("crn")
            crn_testkit.data_hashes = crn_testkit.data_hashes.slice(length=0)
            crn_testkit.fake_run()
            self.backend.create_resolution(
                crn_testkit.source.to_resolution(),
                path=crn_testkit.resolution_path,
            )

            self.backend.insert_source_data(
                crn_testkit.source.resolution_path, crn_testkit.data_hashes
            )
            # Resolution marked as complete
            assert (
                self.backend.get_resolution_stage(crn_testkit.source.resolution_path)
                == UploadStage.COMPLETE
            )

    def test_index_different_resolution_same_hashes(self) -> None:
        """Test that indexing data with the same hashes but different sources works."""
        with self.scenario(self.backend, "bare") as dag_testkit:
            # Prepare original source
            crn_testkit: SourceTestkit = dag_testkit.sources.get("crn").fake_run()
            # Create new source with same hashes
            dh_testkit: SourceTestkit = dag_testkit.sources.get("dh")
            dh_testkit.data_hashes = crn_testkit.data_hashes
            dh_testkit.fake_run()

            # Add original source
            self.backend.create_resolution(
                crn_testkit.source.to_resolution(), path=crn_testkit.resolution_path
            )
            self.backend.insert_source_data(
                crn_testkit.source.resolution_path, crn_testkit.data_hashes
            )
            # Add different source, with same hashes
            self.backend.create_resolution(
                dh_testkit.source.to_resolution(), path=dh_testkit.resolution_path
            )
            self.backend.insert_source_data(
                dh_testkit.source.resolution_path, crn_testkit.data_hashes
            )
            assert self.backend.data.count() == len(crn_testkit.data_hashes)
            assert self.backend.source_resolutions.count() == 2

    def test_insert_model(self) -> None:
        """Test that models can be inserted."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # Test deduper insertion
            models_count = self.backend.models.count()

            dedupe_1_testkit = model_factory(
                name="dedupe_1",
                description="Test deduper 1",
                dag=dag_testkit.dag,
                left_testkit=crn_testkit,
                true_entities=linked.true_entities,
            )
            self.backend.create_resolution(
                resolution=dedupe_1_testkit.fake_run().model.to_resolution(),
                path=dedupe_1_testkit.resolution_path,
            )

            dedupe_2_testkit = model_factory(
                name="dedupe_2",
                description="Test deduper 2",
                dag=dag_testkit.dag,
                left_testkit=dh_testkit,
                true_entities=linked.true_entities,
            )
            self.backend.create_resolution(
                resolution=dedupe_2_testkit.fake_run().model.to_resolution(),
                path=dedupe_2_testkit.resolution_path,
            )

            assert self.backend.models.count() == models_count + 2

            # Test linker insertion
            linker_testkit = model_factory(
                name="link_1",
                description="Test linker 1",
                left_testkit=dedupe_1_testkit,
                right_testkit=dedupe_2_testkit,
                true_entities=linked.true_entities,
            )
            self.backend.create_resolution(
                resolution=linker_testkit.fake_run().model.to_resolution(),
                path=linker_testkit.resolution_path,
            )

            assert self.backend.models.count() == models_count + 3

            # We cannot re-create under the same path
            with pytest.raises(MatchboxResolutionAlreadyExists):
                self.backend.create_resolution(
                    linker_testkit.fake_run().model.to_resolution(),
                    path=linker_testkit.resolution_path,
                )

            assert self.backend.models.count() == models_count + 3

            # Can update model resolution
            old_resolution = linker_testkit.model.to_resolution()
            updated_config = ModelConfig.model_validate(
                old_resolution.config.model_copy(
                    update={
                        "left_query": old_resolution.config.left_query.model_copy(
                            update={"threshold": 99}
                        )
                    }
                )
            )
            updated_resolution = Resolution.model_validate(
                old_resolution.model_copy(
                    update={
                        "description": "updated",
                        "truth": 33,
                        "config": updated_config,
                    }
                )
            )
            self.backend.update_resolution(
                resolution=updated_resolution,
                path=linker_testkit.resolution_path,
            )

            # We can retrieve the updated resolution
            linker_retrieved = self.backend.get_resolution(
                linker_testkit.resolution_path
            )
            assert linker_retrieved.description == "updated"
            assert linker_retrieved.truth == 33
            assert linker_retrieved.config.left_query.threshold == 99

            # We cannot change a model's inputs
            rewired_config = ModelConfig.model_validate(
                old_resolution.config.model_copy(
                    update={
                        "left_query": old_resolution.config.left_query.model_copy(
                            update={"model_resolution": "new_model"}
                        )
                    }
                )
            )
            rewired_resolution = Resolution.model_validate(
                old_resolution.model_copy(update={"config": rewired_config})
            )

            with pytest.raises(MatchboxResolutionUpdateError, match="parents"):
                self.backend.update_resolution(
                    resolution=rewired_resolution,
                    path=linker_testkit.resolution_path,
                )

            # We cannot change model results fingerprint
            with pytest.raises(MatchboxResolutionUpdateError, match="fingerprint"):
                corrupt_resolution = Resolution.model_validate(
                    updated_resolution.model_copy(update={"fingerprint": b"fake"})
                )
                self.backend.update_resolution(
                    resolution=corrupt_resolution,
                    path=linker_testkit.resolution_path,
                )

    def test_model_results_basic(self) -> None:
        """Test that a model's results data can be set and retrieved."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            # Query returns the same results as the testkit, showing
            # that processing was performed accurately.
            # (that we can query from it implies the resolution was correctly
            # marked as complete)
            res = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )
            res_clusters = query_to_cluster_entities(
                data=res,
                keys={crn_testkit.name: "key"},
            )

            identical, report = diff_results(
                expected=naive_crn_testkit.entities,
                actual=res_clusters,
            )

            assert identical, report

            # Retrieve
            pre_results = self.backend.get_model_data(
                path=naive_crn_testkit.resolution_path
            )

            assert isinstance(pre_results, pa.Table)
            assert len(pre_results) > 0

            # Validate IDs
            self.backend.validate_ids(ids=pre_results["left_id"].to_pylist())
            self.backend.validate_ids(ids=pre_results["right_id"].to_pylist())

            # Cannot set new results
            with pytest.raises(MatchboxResolutionExistingData):
                self.backend.insert_model_data(
                    path=naive_crn_testkit.resolution_path,
                    results=naive_crn_testkit.probabilities.to_arrow(),
                )

            # Retrieve again
            post_results = self.backend.get_model_data(
                path=naive_crn_testkit.resolution_path
            )

            # Check difference
            assert pre_results == post_results

    def test_model_results_probabilistic(self) -> None:
        """Test that a probabilistic model's results data can be set and retrieved."""
        with self.scenario(self.backend, "probabilistic_dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            prob_crn_testkit = dag_testkit.models.get("probabilistic_test_crn")

            # Query returns the same results as the testkit, showing
            # that processing was performed accurately
            res = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=prob_crn_testkit.resolution_path,
            )
            res_clusters = query_to_cluster_entities(
                data=res,
                keys={crn_testkit.name: "key"},
            )

            identical, report = diff_results(
                expected=prob_crn_testkit.entities,
                actual=res_clusters,
            )
            assert identical, report

            # Retrieve
            pre_results = self.backend.get_model_data(
                path=prob_crn_testkit.resolution_path
            )

            assert isinstance(pre_results, pa.Table)
            assert len(pre_results) > 0

            # Validate IDs
            self.backend.validate_ids(ids=pre_results["left_id"].to_pylist())
            self.backend.validate_ids(ids=pre_results["right_id"].to_pylist())

    def test_model_results_empty(self) -> None:
        """Can insert and retrieve empty model results"""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            model_testkit = model_factory(
                model_type="deduper", dag=crn_testkit.source.dag
            )
            model_testkit.probabilities = model_testkit.probabilities.head(0)
            model_testkit.fake_run()

            self.backend.create_resolution(
                model_testkit.model.to_resolution(), path=model_testkit.resolution_path
            )

            self.backend.insert_model_data(
                path=model_testkit.model.resolution_path,
                results=model_testkit.model.results.probabilities.to_arrow(),
            )

            # Querying from deduper with no results is the same as querying from source
            # (That we can query also implies that resolution marked as complete)
            source_query = self.backend.query(source=crn_testkit.resolution_path)
            dedupe_query = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=model_testkit.resolution_path,
            )

            assert source_query == dedupe_query

    def test_model_results_shared_clusters(self) -> None:
        """Test that model results data can be inserted when clusters are shared."""
        with self.scenario(self.backend, "convergent") as dag_testkit:
            for model_testkit in dag_testkit.models.values():
                self.backend.create_resolution(
                    path=model_testkit.resolution_path,
                    resolution=model_testkit.fake_run().model.to_resolution(),
                )
                self.backend.insert_model_data(
                    path=model_testkit.resolution_path,
                    results=model_testkit.probabilities.to_arrow(),
                )

    # Data management

    def test_validate_ids(self) -> None:
        """Test validating data IDs."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            df_crn = self.backend.query(
                source=crn_testkit.source.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )

            ids = df_crn["id"].to_pylist()
            assert len(ids) > 0
            self.backend.validate_ids(ids=ids)

            with pytest.raises(MatchboxDataNotFound):
                self.backend.validate_ids(ids=[-6])

    def test_clear(self) -> None:
        """Test deleting all rows in the database."""
        with self.scenario(self.backend, "dedupe"):
            assert self.backend.sources.count() > 0
            assert self.backend.data.count() > 0
            assert self.backend.models.count() > 0
            assert self.backend.clusters.count() > 0
            assert self.backend.creates.count() > 0
            assert self.backend.merges.count() > 0
            assert self.backend.proposes.count() > 0

            self.backend.clear(certain=True)

            assert self.backend.sources.count() == 0
            assert self.backend.data.count() == 0
            assert self.backend.models.count() == 0
            assert self.backend.clusters.count() == 0
            assert self.backend.creates.count() == 0
            assert self.backend.merges.count() == 0
            assert self.backend.proposes.count() == 0

    def test_clear_and_restore(self) -> None:
        """Test that clearing and restoring the database works."""
        with self.scenario(self.backend, "link") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            count_funcs = [
                self.backend.sources.count,
                self.backend.models.count,
                self.backend.data.count,
                self.backend.clusters.count,
                self.backend.merges.count,
                self.backend.creates.count,
                self.backend.proposes.count,
            ]

            def get_counts() -> list[int]:
                return [f() for f in count_funcs]

            # Verify we have data
            pre_dump_counts = get_counts()
            assert all(count > 0 for count in pre_dump_counts)

            # Get some specific IDs to verify they're restored properly
            df_crn_before = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )
            sample_ids_before = df_crn_before["id"].to_pylist()[:5]  # Take first 5 IDs

            # Dump the database
            snapshot = self.backend.dump()

        with self.scenario(self.backend, "bare") as _:
            # Verify counts match pre-dump state
            assert all(c == 0 for c in get_counts())

            # Restore from snapshot
            self.backend.restore(snapshot)

            # Verify counts match pre-dump state
            assert get_counts() == pre_dump_counts

            # Verify specific data was restored correctly
            df_crn_after = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )
            sample_ids_after = df_crn_after["id"].to_pylist()[:5]  # Take first 5 IDs

            # The same IDs should be present after restoration
            assert set(sample_ids_before) == set(sample_ids_after)

            # Test that restoring also clears the database
            self.backend.restore(snapshot)

            # Verify counts still match
            assert get_counts() == pre_dump_counts

    # User management

    def test_login(self) -> None:
        """Can swap user name with user ID."""
        with self.scenario(self.backend, "bare") as _:
            alice_id = self.backend.login("alice")
            assert alice_id == self.backend.login("alice")
            assert alice_id != self.backend.login("bob")

    # Evaluation management

    def test_insert_and_get_judgement(self) -> None:
        """Can insert and retrieve judgements."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            # To begin with, no judgements to retrieve
            judgements, expansion = self.backend.get_judgements()
            assert len(judgements) == len(expansion) == 0

            # Do some queries to find real source cluster IDs
            deduped_query = pl.from_arrow(
                self.backend.query(
                    source=crn_testkit.resolution_path,
                    point_of_truth=naive_crn_testkit.resolution_path,
                )
            )
            unique_ids = deduped_query["id"].unique()
            all_leaves = pl.from_arrow(
                self.backend.query(source=crn_testkit.resolution_path)
            )

            def get_leaf_ids(cluster_id: int) -> list[int]:
                return (
                    deduped_query.filter(pl.col("id") == cluster_id)
                    .join(all_leaves, on="key", suffix="_leaf")["id_leaf"]
                    .to_list()
                )

            alice_id = self.backend.login("alice")

            original_cluster_num = self.backend.clusters.count()

            # Can endorse the same cluster that is shown
            clust1_leaves = get_leaf_ids(unique_ids[0])
            self.backend.insert_judgement(
                judgement=Judgement(
                    user_id=alice_id,
                    shown=unique_ids[0],
                    endorsed=[clust1_leaves],
                ),
            )
            # Can send redundant data
            self.backend.insert_judgement(
                judgement=Judgement(
                    user_id=alice_id,
                    shown=unique_ids[0],
                    endorsed=[clust1_leaves],
                ),
            )
            assert self.backend.clusters.count() == original_cluster_num

            # Now split a cluster
            clust2_leaves = get_leaf_ids(unique_ids[1])
            self.backend.insert_judgement(
                judgement=Judgement(
                    user_id=alice_id,
                    shown=unique_ids[1],
                    endorsed=[clust2_leaves[:1], clust2_leaves[1:]],
                ),
            )
            # Now, two new clusters should have been created
            assert self.backend.clusters.count() == original_cluster_num + 2

            # Let's check failures
            # First, confirm that the following leaves don't exist
            fake_leaves = [10000, 10001]
            with pytest.raises(MatchboxDataNotFound):
                self.backend.validate_ids(fake_leaves)
            # Now, let's test an exception is raised
            with pytest.raises(MatchboxDataNotFound):
                self.backend.insert_judgement(
                    judgement=Judgement(
                        user_id=alice_id, shown=unique_ids[0], endorsed=[fake_leaves]
                    ),
                )

            # Now, let's try to get the judgements back
            # Data gets back in the right shape
            judgements, expansion = self.backend.get_judgements()
            assert judgements.schema.equals(SCHEMA_JUDGEMENTS)
            assert expansion.schema.equals(SCHEMA_CLUSTER_EXPANSION)
            # Only one user ID was used
            assert judgements["user_id"].unique().to_pylist() == [alice_id]
            # The first shown cluster is repeated because we judged it twice
            # The second shown cluster is repeated because we split it (see above)
            assert sorted(judgements["shown"].to_pylist()) == sorted(
                [unique_ids[0], unique_ids[0], unique_ids[1], unique_ids[1]]
            )
            # On the other hand, the root-leaf mapping table has no duplicates
            assert len(expansion) == 4  # 2 shown clusters + 2 new endorsed clusters

            # Let's massage tables into a root-leaf dict for all endorsed clusters
            endorsed_dict = dict(
                pl.from_arrow(judgements)
                .join(pl.from_arrow(expansion), left_on="endorsed", right_on="root")
                .select(["endorsed", "leaves"])
                .rows()
            )

            # The root we know about has the leaves we expect
            assert set(endorsed_dict[unique_ids[0]]) == set(clust1_leaves)
            # Other than the root we know about, there are two new ones
            assert len(set(endorsed_dict.keys())) == 3
            # The other two sets of leaves are there too
            assert set(map(frozenset, endorsed_dict.values())) == set(
                map(frozenset, [clust1_leaves, clust2_leaves[:1], clust2_leaves[1:]])
            )

    def test_compare_models_fails(self) -> None:
        """Model comparison errors with no judgement data."""
        with (
            self.scenario(self.backend, "bare"),
            pytest.raises(MatchboxNoJudgements),
        ):
            self.backend.compare_models([])

    def test_compare_models(self) -> None:
        """Can compute precision and recall for list of models."""
        with self.scenario(self.backend, "alt_dedupe") as dag_testkit:
            user_id = self.backend.login("alice")

            model_names = [
                model.resolution_path for model in dag_testkit.models.values()
            ]

            root_leaves = (
                pl.from_arrow(
                    self.backend.sample_for_eval(
                        n=10,
                        path=model_names[0],
                        user_id=user_id,
                    )
                )
                .select(["root", "leaf"])
                .unique()
                .group_by("root")
                .agg("leaf")
            )
            for row in root_leaves.rows(named=True):
                self.backend.insert_judgement(
                    judgement=Judgement(
                        user_id=user_id, shown=row["root"], endorsed=[row["leaf"]]
                    )
                )

            pr = self.backend.compare_models(model_names)
            # Precision must be 1 for both as the second model is like the first
            # but more conservative
            assert pr[model_names[0]][0] == pr[model_names[1]][0] == 1
            # Recall must be 1 for the first model and lower for the second
            assert pr[model_names[0]][1] == 1
            assert pr[model_names[1]][1] < 1

    def test_sample_for_eval(self) -> None:
        """Can extract samples for a user and a resolution."""

        # Missing resolution raises error
        with (
            self.scenario(self.backend, "bare"),
            pytest.raises(MatchboxResolutionNotFoundError, match="naive_test_crn"),
        ):
            user_id = self.backend.login("alice")
            self.backend.sample_for_eval(
                n=10,
                path=ResolutionPath(
                    collection="collection", run=1, name="naive_test_crn"
                ),
                user_id=user_id,
            )

        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            user_id = self.backend.login("alice")

            # Source clusters should not be returned
            # So if we sample from a source resolution, we get nothing
            user_id = self.backend.login("alice")
            samples_source = self.backend.sample_for_eval(
                n=10, path=crn_testkit.resolution_path, user_id=user_id
            )
            assert len(samples_source) == 0

            # We now look at more interesting cases
            # Query backend to form expectations
            resolution_clusters = pl.from_arrow(
                self.backend.query(
                    source=crn_testkit.resolution_path,
                    point_of_truth=naive_crn_testkit.resolution_path,
                )
            )
            source_clusters = pl.from_arrow(
                self.backend.query(source=crn_testkit.resolution_path)
            )
            # We can request more than available
            assert len(resolution_clusters["id"].unique()) < 99

            samples_99 = self.backend.sample_for_eval(
                n=99, path=naive_crn_testkit.resolution_path, user_id=user_id
            )

            assert samples_99.schema.equals(SCHEMA_EVAL_SAMPLES)

            # We can reconstruct the expected sample from resolution and source queries
            expected_sample = (
                resolution_clusters.join(source_clusters, on="key", suffix="_source")
                .rename({"id": "root", "id_source": "leaf"})
                .with_columns(pl.lit("crn").alias("source"))
            )

            assert_frame_equal(
                pl.from_arrow(samples_99),
                expected_sample,
                check_row_order=False,
                check_column_order=False,
                check_dtypes=False,
            )

            # We can request less than available
            assert len(resolution_clusters["id"].unique()) > 5
            samples_5 = self.backend.sample_for_eval(
                n=5, path=naive_crn_testkit.resolution_path, user_id=user_id
            )
            assert len(samples_5["root"].unique()) == 5

            # If user has recent judgements, exclude clusters
            first_cluster_id = resolution_clusters["id"][0]
            first_cluster = resolution_clusters.filter(pl.col("id") == first_cluster_id)
            first_cluster_leaves = first_cluster.join(
                source_clusters, on="key", suffix="_source"
            )["id_source"].to_list()

            self.backend.insert_judgement(
                judgement=Judgement(
                    user_id=user_id,
                    shown=first_cluster_id,
                    endorsed=[first_cluster_leaves],
                ),
            )

            samples_without_cluster = self.backend.sample_for_eval(
                n=99, path=naive_crn_testkit.resolution_path, user_id=user_id
            )
            # Compared to the first query, we should have one fewer cluster
            assert len(samples_99["root"].unique()) - 1 == len(
                samples_without_cluster["root"].unique()
            )
            # And that cluster is the one on which the judgement is based
            assert first_cluster_id in samples_99["root"].to_pylist()
            assert first_cluster_id not in samples_without_cluster["root"].to_pylist()

            # If a user has judged all available clusters, nothing is returned
            for cluster_id in resolution_clusters["id"].unique():
                cluster = resolution_clusters.filter(pl.col("id") == cluster_id)
                cluster_leaves = cluster.join(
                    source_clusters, on="key", suffix="_source"
                )["id_source"].to_list()

                self.backend.insert_judgement(
                    judgement=Judgement(
                        user_id=user_id,
                        shown=cluster_id,
                        endorsed=[cluster_leaves],
                    ),
                )

            samples_all_done = self.backend.sample_for_eval(
                n=99, path=naive_crn_testkit.resolution_path, user_id=user_id
            )
            assert len(samples_all_done) == 0



import logging
from collections.abc import Generator

import polars as pl
import pytest
from httpx import Client
from polars.testing import assert_frame_equal
from sqlalchemy import Engine, text

from matchbox.client.dags import DAG
from matchbox.client.locations import RelationalDBLocation
from matchbox.client.models.dedupers import NaiveDeduper
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.common.exceptions import MatchboxResolutionNotFoundError
from matchbox.common.factories.sources import (
    FeatureConfig,
    SourceTestkitParameters,
    SuffixRule,
    linked_sources_factory,
)


@pytest.mark.docker
class TestE2EPipelineBuilder:
    """End to end tests for DAG pipeline functionality."""

    def _clean_company_name(self, column: str) -> str:
        """Generate cleaning SQL for a company name column.

        Removes company suffixes (Ltd, Limited) and normalises whitespace
        from the company_name field.
        """

        return f"""
            trim(
                regexp_replace(
                    regexp_replace(
                        {column},
                        ' (Ltd|Limited)$',
                        '',
                        'g'
                    ),
                    '\\s+',
                    ' ',
                    'g'
                )
            )
        """

    @pytest.fixture(autouse=True, scope="function")
    def setup_environment(
        self,
        matchbox_client: Client,
        postgres_warehouse: Engine,
    ) -> Generator[None, None, None]:
        """Set up warehouse and database using fixtures."""
        # Persist shared setup for use in the test body
        n_true_entities = 10  # Keep it small for simplicity
        self.warehouse_engine = postgres_warehouse

        # Create simple feature configurations - just two sources
        features = {
            "company_name": FeatureConfig(
                name="company_name",
                base_generator="company",
            ),
            "registration_id": FeatureConfig(
                name="registration_id",
                base_generator="bothify",
                parameters=(("text", "REG-###-???"),),
            ),
        }

        # Create two simple sources that can be linked
        source_parameters = (
            SourceTestkitParameters(
                name="source_a",
                engine=postgres_warehouse,
                features=(
                    features["company_name"].add_variations(
                        SuffixRule(suffix=" Ltd"),
                        SuffixRule(suffix=" Limited"),
                    ),
                    features["registration_id"],
                ),
                n_true_entities=n_true_entities,
                repetition=0,  # No duplicates
            ),
            SourceTestkitParameters(
                name="source_b",
                engine=postgres_warehouse,
                features=(
                    features["company_name"],
                    features["registration_id"],
                ),
                n_true_entities=n_true_entities // 2,  # Half overlap
                repetition=1,  # Duplicate all rows for deduplication testing
            ),
        )

        # Create linked sources testkit
        self.linked_testkit = linked_sources_factory(
            source_parameters=source_parameters,
            seed=42,
        )

        # Setup - Create tables in warehouse
        for source_testkit in self.linked_testkit.sources.values():
            source_testkit.write_to_location()

        # Clear matchbox database before test
        response = matchbox_client.delete("/database", params={"certain": "true"})
        assert response.status_code == 200, "Failed to clear matchbox database"

        yield

        # Teardown
        with postgres_warehouse.connect() as conn:
            for source_name in self.linked_testkit.sources:
                conn.execute(text(f"DROP TABLE IF EXISTS {source_name};"))
            conn.commit()

        response = matchbox_client.delete("/database", params={"certain": "true"})
        assert response.status_code == 200, "Failed to clear matchbox database"

    def test_dag_pipeline_creation_and_rerun(self) -> None:
        """Test DAG API with simple two-source pipeline

        Rerun to test overwriting.
        """

        # === SETUP PHASE ===
        dw_loc = RelationalDBLocation(name="dbname").set_client(self.warehouse_engine)

        dag = DAG("companies").new_run()

        # Create source configs
        source_a = dag.source(
            location=dw_loc,
            name="source_a",
            extract_transform="""
                select
                    key::text as id,
                    company_name,
                    registration_id
                from
                    source_a;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name", "registration_id"],
        )

        source_b = dag.source(
            location=dw_loc,
            name="source_b",
            extract_transform="""
                select
                    key::text as id,
                    company_name,
                    registration_id
                from
                    source_b;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name", "registration_id"],
        )

        # Dedupe steps
        dedupe_a = source_a.query(
            cleaning={
                "company_name": self._clean_company_name(source_a.f("company_name")),
                "registration_id": source_a.f("registration_id"),
            },
        ).deduper(
            name="dedupe_source_a",
            description="Deduplicate source A",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": ["registration_id"]},
        )

        dedupe_b = source_b.query(
            cleaning={
                "company_name": self._clean_company_name(source_b.f("company_name")),
                "registration_id": source_b.f("registration_id"),
            }
        ).deduper(
            name="dedupe_source_b",
            description="Deduplicate source B",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": ["registration_id"]},
        )

        # Link deduplicated sources A and B
        link_a_b = dedupe_a.query(
            source_a,
            cleaning={
                "company_name": self._clean_company_name(source_a.f("company_name")),
                "registration_id": source_a.f("registration_id"),
            },
        ).linker(
            dedupe_b.query(
                source_b,
                cleaning={
                    "company_name": self._clean_company_name(
                        source_b.f("company_name")
                    ),
                    "registration_id": source_b.f("registration_id"),
                },
            ),
            name="final",
            description="Link sources A and B on registration ID",
            model_class=DeterministicLinker,
            model_settings={"comparisons": ["l.registration_id = r.registration_id"]},
        )

        # === FIRST RUN ===
        logging.info("Running DAG for the first time")
        dag.run_and_sync()

        # Update metadata of one node, will check later
        link_a_b.description = "Updated description"
        link_a_b.sync()

        # Basic verification - we have some linked results and can retrieve them
        final_df = link_a_b.query(source_a, source_b).run()

        # # Should have linked results
        assert len(final_df) > 0, "Expected some results from first run"
        assert final_df["id"].n_unique() == len(self.linked_testkit.true_entities)

        first_run_entities = final_df["id"].n_unique()
        logging.info(f"First run produced {first_run_entities} unique entities")

        # Lookup works too
        test_key = final_df.filter(
            pl.col(source_b.f(source_b.config.key_field.name)).is_not_null()
        )[source_b.f(source_b.config.key_field.name)][0]

        matches = dag.lookup_key(
            from_source=source_b.name,
            to_sources=[source_a.name],
            key=test_key,
        )
        assert len(matches[source_a.name]) >= 1

        # Can retrieve whole lookup
        dag1_lookup = dag.extract_lookup()

        # Set as new default
        dag.set_default()

        # === SECOND RUN ===

        logging.info("Running DAG again to test downloading and using the default")

        # Load default
        reconstructed_dag = DAG("companies").load_default()
        assert reconstructed_dag.run == dag.run
        # Previous update was effective
        assert reconstructed_dag.get_model("final").description == "Updated description"

        # Can directly read data from default
        assert matches == reconstructed_dag.lookup_key(
            from_source=source_b.name,
            to_sources=[source_a.name],
            key=test_key,
        )

        # Start a new run
        rerun_dag = reconstructed_dag.set_client(self.warehouse_engine).new_run()
        assert rerun_dag.run != dag.run
        rerun_dag.run_and_sync()

        # The lookup is identical
        assert_frame_equal(
            pl.from_arrow(rerun_dag.extract_lookup()),
            pl.from_arrow(dag1_lookup),
            check_column_order=False,
            check_row_order=False,
        )

        # Load pending to check we can
        pending_dag: DAG = (
            DAG("companies").load_pending().set_client(self.warehouse_engine)
        )
        assert pending_dag.run == rerun_dag.run

        logging.info("DAG pipeline test completed successfully!")

        # Possible to overwrite node locally
        # (following source has one fewer field)
        source_a = pending_dag.source(
            location=dw_loc,
            name="source_a",
            extract_transform="""
                select
                    key::text as id,
                    company_name
                from
                    source_a;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name"],
        )

        source_a.run()
        # Possible to overwrite one node on server
        source_a.sync()
        # This will cause downstream queries to fail
        with pytest.raises(MatchboxResolutionNotFoundError):
            pending_dag.get_model("final").query(source_a).run()



#!/usr/bin/env python3
"""Script to run Textual evaluation app with scenario data via CLI."""

import logging
import tempfile
from collections.abc import Generator
from contextlib import contextmanager, suppress
from pathlib import Path
from typing import Annotated

import typer
from sqlalchemy import create_engine

from matchbox.client._settings import settings
from matchbox.client.cli.eval import run
from matchbox.common.factories import models as _  # noqa:F401
from matchbox.common.factories.scenarios import SCENARIO_REGISTRY, setup_scenario
from matchbox.server.base import (
    MatchboxDBAdapter,
    MatchboxServerSettings,
    get_backend_settings,
    settings_to_backend,
)

# Set up logger for this script
logger = logging.getLogger(__name__)


def _get_backend() -> MatchboxDBAdapter:
    """Instantiates a backend class based on the contents of .env."""
    SettingsClass = get_backend_settings(MatchboxServerSettings().backend_type)
    settings = SettingsClass()
    return settings_to_backend(settings)


@contextmanager
def scenario_setup(scenario_name: str) -> Generator[dict[str, object], None, None]:
    """Context manager that sets up scenario data and yields CLI parameters."""

    # Create temporary SQLite database
    with tempfile.NamedTemporaryFile(suffix=".db", delete=False) as tmp:
        tmp_path = Path(tmp.name)

    try:
        warehouse_engine = create_engine(f"sqlite:///{tmp_path}")

        # Configure client to use our SQLite warehouse
        settings.default_warehouse = str(warehouse_engine.url)

        backend = _get_backend()
        backend.clear(certain=True)

        logger.info(f"Setting up {scenario_name} scenario...")

        with setup_scenario(
            backend=backend,
            scenario_type=scenario_name,
            warehouse=warehouse_engine,
            n_entities=10,
            seed=42,
        ) as dag:
            logger.info("Scenario ready! Starting Textual evaluate app via CLI...")

            # Select the appropriate resolution for each scenario
            match scenario_name:
                case "bare" | "index" | "convergent":
                    raise RuntimeError("Scenario has nothing to evaluate.")
                case "dedupe":
                    resolution = dag.dag.nodes["naive_test_crn"]
                case "probabilistic_dedupe":
                    resolution = dag.dag.nodes["probabilistic_test_crn"]
                case "link":
                    resolution = dag.dag.nodes["final_join"]
                case "alt_dedupe":
                    resolution = dag.dag.nodes["dedupe_foo_a"]
                case "mega":
                    resolution = dag.dag.nodes["mega_product_linker"]
                case _:
                    raise ValueError(f"Unknown scenario: {scenario_name}")

            # Yield CLI parameters instead of app instance
            yield {
                "collection": resolution.resolution_path.collection,
                "resolution": resolution.resolution_path.name,
                "pending": True,
                "warehouse": str(warehouse_engine.url),
            }

    finally:
        # Clean up
        with suppress(Exception):
            backend.clear(certain=True)
        with suppress(Exception):
            tmp_path.unlink(missing_ok=True)


def main(
    scenario: Annotated[
        str,
        typer.Argument(
            help=f"Scenario type. Available: {', '.join(SCENARIO_REGISTRY.keys())}"
        ),
    ] = None,
    log_file: Annotated[
        str | None,
        typer.Option(
            "--log",
            help="Log file path to redirect all logging output (keeps UI clean)",
        ),
    ] = None,
) -> None:
    """Run the scenario-based evaluation app via CLI."""
    # Set up basic logging for this script
    logging.basicConfig(level=logging.INFO, format="%(message)s")

    # If no scenario provided, list available scenarios
    if scenario is None or scenario == "":
        available_scenarios = list(SCENARIO_REGISTRY.keys())
        logger.info("Available scenarios:")
        for i, name in enumerate(available_scenarios, 1):
            logger.info(f"  {i}. {name}")
        logger.info(
            "\nUsage: uv run python test/scripts/eval.py <scenario> [--log file.log]"
        )
        raise typer.Exit(0)

    if scenario not in SCENARIO_REGISTRY:
        available = ", ".join(SCENARIO_REGISTRY.keys())
        logger.error(f"Unknown scenario: {scenario}")
        logger.info(f"Available scenarios: {available}")
        raise typer.Exit(1)

    try:
        with scenario_setup(scenario) as cli_params:
            # Call CLI evaluation command directly
            try:
                run.evaluate(
                    collection=cli_params["collection"],
                    resolution=cli_params["resolution"],
                    pending=cli_params["pending"],
                    warehouse=cli_params["warehouse"],
                    user=None,
                    log_file=log_file if log_file and log_file.strip() else None,
                )
            except KeyboardInterrupt as e:
                logger.info("\nKeyboard interrupt received, stopping...")
                raise typer.Exit(0) from e

    except KeyboardInterrupt as e:
        logger.info("\nExiting...")
        raise typer.Exit(0) from e
    except Exception as e:
        logger.error(f"Error: {e}")
        raise typer.Exit(1) from e


if __name__ == "__main__":
    typer.run(main)



"""Create tables given schema.

Revision ID: 1907c34cfa1f
Revises: 40a8e5ed48f2
"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "1907c34cfa1f"
down_revision: str | None = "40a8e5ed48f2"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.create_table(
        "clusters",
        sa.Column("cluster_id", sa.BIGINT(), nullable=False),
        sa.Column("cluster_hash", postgresql.BYTEA(), nullable=False),
        sa.PrimaryKeyConstraint("cluster_id"),
        sa.UniqueConstraint("cluster_hash", name="clusters_hash_key"),
        schema="mb",
    )
    op.create_table(
        "resolutions",
        sa.Column("resolution_id", sa.BIGINT(), nullable=False),
        sa.Column("resolution_hash", postgresql.BYTEA(), nullable=False),
        sa.Column("type", sa.TEXT(), nullable=False),
        sa.Column("name", sa.TEXT(), nullable=False),
        sa.Column("description", sa.TEXT(), nullable=True),
        sa.Column("truth", sa.SMALLINT(), nullable=True),
        sa.CheckConstraint(
            "type IN ('model', 'dataset', 'human')", name="resolution_type_constraints"
        ),
        sa.PrimaryKeyConstraint("resolution_id"),
        sa.UniqueConstraint("name", name="resolutions_name_key"),
        sa.UniqueConstraint("resolution_hash", name="resolutions_hash_key"),
        schema="mb",
    )
    op.create_table(
        "contains",
        sa.Column("parent", sa.BIGINT(), nullable=False),
        sa.Column("child", sa.BIGINT(), nullable=False),
        sa.CheckConstraint("parent != child", name="no_self_containment"),
        sa.ForeignKeyConstraint(
            ["child"], ["mb.clusters.cluster_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["parent"], ["mb.clusters.cluster_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("parent", "child"),
        schema="mb",
    )
    op.create_index(
        "ix_contains_child_parent",
        "contains",
        ["child", "parent"],
        unique=False,
        schema="mb",
    )
    op.create_index(
        "ix_contains_parent_child",
        "contains",
        ["parent", "child"],
        unique=False,
        schema="mb",
    )
    op.create_table(
        "probabilities",
        sa.Column("resolution", sa.BIGINT(), nullable=False),
        sa.Column("cluster", sa.BIGINT(), nullable=False),
        sa.Column("probability", sa.SMALLINT(), nullable=False),
        sa.CheckConstraint("probability BETWEEN 0 AND 100", name="valid_probability"),
        sa.ForeignKeyConstraint(
            ["cluster"], ["mb.clusters.cluster_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["resolution"], ["mb.resolutions.resolution_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("resolution", "cluster"),
        schema="mb",
    )
    op.create_table(
        "resolution_from",
        sa.Column("parent", sa.BIGINT(), nullable=False),
        sa.Column("child", sa.BIGINT(), nullable=False),
        sa.Column("level", sa.INTEGER(), nullable=False),
        sa.Column("truth_cache", sa.SMALLINT(), nullable=True),
        sa.CheckConstraint("level > 0", name="positive_level"),
        sa.CheckConstraint("parent != child", name="no_self_reference"),
        sa.ForeignKeyConstraint(
            ["child"], ["mb.resolutions.resolution_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["parent"], ["mb.resolutions.resolution_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("parent", "child"),
        schema="mb",
    )
    op.create_table(
        "sources",
        sa.Column(
            "source_id", sa.BIGINT(), sa.Identity(always=False, start=1), nullable=False
        ),
        sa.Column("resolution_id", sa.BIGINT(), nullable=True),
        sa.Column("resolution_name", sa.TEXT(), nullable=False),
        sa.Column("full_name", sa.TEXT(), nullable=False),
        sa.Column("warehouse_hash", postgresql.BYTEA(), nullable=False),
        sa.Column("db_pk", sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(
            ["resolution_id"], ["mb.resolutions.resolution_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("source_id"),
        sa.UniqueConstraint(
            "full_name", "warehouse_hash", name="unique_source_address"
        ),
        schema="mb",
    )
    op.create_table(
        "cluster_source_pks",
        sa.Column("pk_id", sa.BIGINT(), nullable=False),
        sa.Column("cluster_id", sa.BIGINT(), nullable=False),
        sa.Column("source_id", sa.BIGINT(), nullable=False),
        sa.Column("source_pk", sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(
            ["cluster_id"], ["mb.clusters.cluster_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["source_id"], ["mb.sources.source_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("pk_id"),
        sa.UniqueConstraint("pk_id", "source_id", name="unique_pk_source"),
        schema="mb",
    )
    op.create_index(
        "ix_cluster_source_pks_cluster_id",
        "cluster_source_pks",
        ["cluster_id"],
        unique=False,
        schema="mb",
    )
    op.create_index(
        "ix_cluster_source_pks_source_pk",
        "cluster_source_pks",
        ["source_pk"],
        unique=False,
        schema="mb",
    )
    op.create_table(
        "source_columns",
        sa.Column("column_id", sa.BIGINT(), nullable=False),
        sa.Column("source_id", sa.BIGINT(), nullable=False),
        sa.Column("column_index", sa.INTEGER(), nullable=False),
        sa.Column("column_name", sa.TEXT(), nullable=False),
        sa.Column("column_type", sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(
            ["source_id"], ["mb.sources.source_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("column_id"),
        sa.UniqueConstraint("source_id", "column_index", name="unique_column_index"),
        schema="mb",
    )
    op.create_index(
        "ix_source_columns_source_id",
        "source_columns",
        ["source_id"],
        unique=False,
        schema="mb",
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_index(
        "ix_source_columns_source_id", table_name="source_columns", schema="mb"
    )
    op.drop_table("source_columns", schema="mb")
    op.drop_index(
        "ix_cluster_source_pks_source_pk", table_name="cluster_source_pks", schema="mb"
    )
    op.drop_index(
        "ix_cluster_source_pks_cluster_id", table_name="cluster_source_pks", schema="mb"
    )
    op.drop_table("cluster_source_pks", schema="mb")
    op.drop_table("sources", schema="mb")
    op.drop_table("resolution_from", schema="mb")
    op.drop_table("probabilities", schema="mb")
    op.drop_index("ix_contains_parent_child", table_name="contains", schema="mb")
    op.drop_index("ix_contains_child_parent", table_name="contains", schema="mb")
    op.drop_table("contains", schema="mb")
    op.drop_table("resolutions", schema="mb")
    op.drop_table("clusters", schema="mb")



"""Logging utilities."""

import importlib.metadata
import logging
from typing import Any, Final, Literal

from rich.console import Console
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

_PLUGINS = None


def get_formatter() -> logging.Formatter:
    """Retrieve plugin registered in 'matchbox.logging' entry point, or fallback."""
    global _PLUGINS
    if _PLUGINS is None:
        _PLUGINS = []
        for ep in importlib.metadata.entry_points(group="matchbox.logging"):
            try:
                _PLUGINS.append(ep.load()())
            except Exception as e:  # noqa: BLE001
                logger.warning(f"Failed to load logging plugin: {e}")

    if _PLUGINS:
        return _PLUGINS[0]

    return logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")


LogLevelType = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
"""Type for all Python log levels."""


class PrefixedLoggerAdapter(logging.LoggerAdapter):
    """A logger adapter that supports adding a prefix enclosed in square brackets.

    This adapter allows passing an optional prefix parameter to any logging call
    without modifying the underlying logger.
    """

    def process(
        self,
        msg: Any,  # noqa: ANN401
        kwargs: dict[str, Any],  # noqa: ANN401
    ) -> tuple[Any, dict[str, Any]]:  # noqa: ANN401
        """Process the log message, adding a prefix if provided.

        Args:
            msg: The log message
            kwargs: Additional arguments to the logging method

        Returns:
            Tuple of (modified_message, modified_kwargs)
        """
        prefix = kwargs.pop("prefix", None)

        if prefix:
            msg = f"[{prefix}] {msg}"

        return msg, kwargs


logger: Final[PrefixedLoggerAdapter] = PrefixedLoggerAdapter(
    logging.getLogger("matchbox"), {}
)
"""Logger for Matchbox.

Used for all logging in the Matchbox library.

Allows passing a prefix to any logging call.

Examples:
    ```python
    log_prefix = f"Model {name}"
    logger.debug("Inserting metadata", prefix=log_prefix)
    logger.debug("Inserting data", prefix=log_prefix)
    logger.info("Insert successful", prefix=log_prefix)
    ```
"""

console: Final[Console] = Console()
"""Console for Matchbox.

Used for any CLI utilities in the Matchbox library.
"""


def build_progress_bar(console_: Console | None = None) -> Progress:
    """Create a progress bar."""
    if console_ is None:
        console_ = console

    return Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
        console=console_,
    )



"""Integration tests for all methodology classes."""

from collections.abc import Callable
from typing import Any

import polars as pl
import pytest
from sqlalchemy import Engine

# Import configurator functions from methodology tests
from test.client.models.methodologies.test_dedupers_deterministic import DEDUPERS
from test.client.models.methodologies.test_linkers_deterministic import (
    configure_deterministic_linker,
    configure_splink_linker,
    configure_weighted_deterministic_linker,
)

from matchbox.client.models.dedupers.base import Deduper
from matchbox.client.models.linkers.base import Linker
from matchbox.client.models.linkers.deterministic import DeterministicLinker
from matchbox.client.models.linkers.splinklinker import SplinkLinker
from matchbox.client.models.linkers.weighteddeterministic import (
    WeightedDeterministicLinker,
)
from matchbox.common.factories.scenarios import setup_scenario
from matchbox.common.factories.sources import SourceTestkit
from matchbox.server.base import MatchboxDBAdapter

# Type aliases for configurator functions
DeduperConfigurator = Callable[[SourceTestkit], dict[str, Any]]
LinkerConfigurator = Callable[[SourceTestkit, SourceTestkit], dict[str, Any]]


@pytest.mark.docker
class TestE2EMethodologyIntegration:
    """Integration tests for all methodology classes with real pipeline execution."""

    @pytest.fixture(autouse=True)
    def setup(
        self,
        matchbox_postgres: MatchboxDBAdapter,
        sqlite_warehouse: Engine,
    ) -> None:
        """Set up scenario system for tests."""
        self.backend = matchbox_postgres
        self.warehouse = sqlite_warehouse

    def _clean_field(self, column: str) -> str:
        """Generate basic cleaning SQL."""
        return f"trim({column})"

    @pytest.mark.parametrize(("Deduper", "configure_deduper"), DEDUPERS)
    def test_deduper_integration(
        self, Deduper: type[Deduper], configure_deduper: DeduperConfigurator
    ) -> None:
        """Test that dedupers work end-to-end."""
        with setup_scenario(self.backend, "index", self.warehouse) as dag_testkit:
            # Get the DAG from the testkit
            dag = dag_testkit.dag

            # Get a source that has duplicates (CRN has repetition in the scenario)
            source_testkit = dag_testkit.sources.get("crn")
            source = dag.get_source("crn")

            # Create settings and deduper
            settings = configure_deduper(source_testkit)

            dedupe_result = source.query(
                cleaning={
                    "company_name": self._clean_field(source.f("company_name")),
                    "crn": source.f("crn"),
                },
            ).deduper(
                name=f"test_{Deduper.__name__}",
                description=f"Integration test for {Deduper.__name__}",
                model_class=Deduper,
                model_settings=settings,
            )

            # Get results
            results = dedupe_result.run()

            # Basic integration checks
            assert results is not None, f"{Deduper.__name__} returned None"
            assert results.probabilities is not None
            assert isinstance(results.probabilities, pl.DataFrame)
            assert len(results.probabilities) >= 0

    @pytest.mark.parametrize(
        ("Linker", "configure_linker"),
        [
            pytest.param(
                DeterministicLinker,
                configure_deterministic_linker,
                id="DeterministicLinker",
            ),
            pytest.param(
                WeightedDeterministicLinker,
                configure_weighted_deterministic_linker,
                id="WeightedDeterministicLinker",
            ),
            pytest.param(
                SplinkLinker,
                configure_splink_linker,
                id="SplinkLinker",
            ),
        ],
    )
    def test_linker_integration(
        self, Linker: type[Linker], configure_linker: LinkerConfigurator
    ) -> None:
        """Test that linkers work end-to-end."""
        with setup_scenario(self.backend, "index", self.warehouse) as dag_testkit:
            # Get the DAG from the testkit
            dag = dag_testkit.dag

            # Get two sources that can be linked
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            crn_source = dag.get_source("crn")
            dh_source = dag.get_source("dh")

            # Create settings and linker
            settings = configure_linker(crn_testkit, dh_testkit)

            link_result = crn_source.query(
                cleaning={
                    "company_name": self._clean_field(crn_source.f("company_name")),
                },
            ).linker(
                dh_source.query(
                    cleaning={
                        "company_name": self._clean_field(dh_source.f("company_name")),
                    },
                ),
                name=f"test_{Linker.__name__}",
                description=f"Integration test for {Linker.__name__}",
                model_class=Linker,
                model_settings=settings,
            )

            # Get results
            results = link_result.run()

            # Basic integration checks
            assert results is not None, f"{Linker.__name__} returned None"
            assert results.probabilities is not None
            assert isinstance(results.probabilities, pl.DataFrame)
            assert len(results.probabilities) >= 0



import polars as pl
import pytest

from matchbox.common.arrow import SCHEMA_CLUSTER_EXPANSION, SCHEMA_JUDGEMENTS
from matchbox.common.eval import Judgement, precision_recall, process_judgements


def test_judgement_validation() -> None:
    """Judgement validates source cluster IDs."""
    with pytest.raises(ValueError):
        Judgement(user_id=1, shown=10, endorsed=[[1, 2, 3], [3, 4, 5]])

    with pytest.raises(ValueError):
        Judgement(user_id=1, shown=10, endorsed=[[1, 2, 3, 1]])

    # Something sensible does work
    Judgement(user_id=1, shown=10, endorsed=[[1, 2, 3], [4, 5]])


def test_precision_recall_fails() -> None:
    """Test instances where PR computation raises."""
    # No judgements
    model = pl.DataFrame([{"root": 12, "leaf": 1}, {"root": 12, "leaf": 2}])
    empty_judgements = pl.DataFrame([], schema=pl.Schema(SCHEMA_JUDGEMENTS))
    empty_expansion = pl.DataFrame([], schema=pl.Schema(SCHEMA_CLUSTER_EXPANSION))

    with pytest.raises(ValueError, match="Judgements data"):
        precision_recall(
            models_root_leaf=[model],
            judgements=empty_judgements,
            expansion=empty_expansion,
        )

    # No model root-leaf
    empty_model = pl.DataFrame([])
    judgements = pl.DataFrame([{"shown": 12, "endorsed": 12}])
    expansion = pl.DataFrame([{"root": 12, "leaves": [1, 2]}])

    with pytest.raises(ValueError, match="Model data"):
        precision_recall(
            models_root_leaf=[empty_model], judgements=judgements, expansion=expansion
        )


def test_precision_recall() -> None:
    """Test calculation of precision and recall from root-leaf tables."""
    # In this test, one-digit cluster IDs are for source clusters.
    # Multiple-digit cluster IDs decompose to source cluster IDs.
    # For example, 123 maps to 1,2,3
    model1 = pl.DataFrame(
        [
            # (1,2,3)
            {"root": 123, "leaf": 1},
            {"root": 123, "leaf": 2},
            {"root": 123, "leaf": 3},
            # (4,5)
            {"root": 45, "leaf": 4},
            {"root": 45, "leaf": 5},
            # (6,7)
            {"root": 67, "leaf": 6},
            {"root": 67, "leaf": 7},
            # (8,9)
            {"root": 89, "leaf": 8},
            {"root": 89, "leaf": 9},
        ]
    )

    model2 = pl.DataFrame(
        [
            # (1,3)
            {"root": 13, "leaf": 1},
            {"root": 13, "leaf": 3},
            # (2)
            {"root": 2, "leaf": 2},
            # (4)
            {"root": 4, "leaf": 4},
            # (5)
            {"root": 5, "leaf": 5},
            # (6,7)
            {"root": 67, "leaf": 6},
            {"root": 67, "leaf": 7},
        ]
    )

    judgement_cluster_expansion = pl.DataFrame(
        [
            {"root": 123, "leaves": [1, 2, 3]},
            {"root": 67, "leaves": [6, 7]},
            {"root": 45, "leaves": [4, 5]},
            {"root": 12, "leaves": [1, 2]},
        ]
    )

    judgements = pl.DataFrame(
        [
            # Ambiguous but more positive than negative
            {"shown": 123, "endorsed": 12},
            {"shown": 123, "endorsed": 3},
            {"shown": 123, "endorsed": 12},
            {"shown": 123, "endorsed": 3},
            {"shown": 123, "endorsed": 1},
            {"shown": 123, "endorsed": 2},
            {"shown": 123, "endorsed": 3},
            # Ambiguous but more negative than positive
            {"shown": 45, "endorsed": 45},
            {"shown": 45, "endorsed": 4},
            {"shown": 45, "endorsed": 4},
            {"shown": 45, "endorsed": 5},
            {"shown": 45, "endorsed": 5},
            # The following neutralise each other
            {"shown": 67, "endorsed": 67},
            {"shown": 67, "endorsed": 6},
            {"shown": 67, "endorsed": 7},
        ]
    )

    pr_scores = precision_recall(
        models_root_leaf=[model1, model2],
        judgements=judgements,
        expansion=judgement_cluster_expansion,
    )

    # Validation pairs: (12) (as (45) is more negative; (67) is neutralised)
    # Model 1 pairs: (12), (13), (23) (as (67) is neutralised; (89) has extra leaves)
    assert pr_scores[0] == (1 / 4, 1)
    # Model 2 pairs: (13) (as (67) is neutralised)
    assert pr_scores[1] == (0, 0)


def test_process_judgements() -> None:
    """Can convert judgements and expansion to pairs, pair counts and set of leaves."""
    # In this test:
    # cluster IDs < 100 are source clusters
    # 1xx clusters IDs are model clusters
    # 2xx clusters IDs are new judgement clusters (splintered model clusters)

    judgement_cluster_expansion = pl.DataFrame(
        [
            # Shown clusters
            {"root": 100, "leaves": [1, 2]},
            {"root": 101, "leaves": [3, 4, 5]},
            {"root": 102, "leaves": [6, 7]},
            {"root": 103, "leaves": [8, 9]},
            {"root": 104, "leaves": [10, 11]},
            {"root": 105, "leaves": [12, 13]},
            {"root": 106, "leaves": [12, 13, 14]},
            # New "splintered" clusters
            {"root": 200, "leaves": [3, 4]},
        ]
    )

    judgements = pl.DataFrame(
        [
            # A) cluster confirmed as is
            {"shown": 100, "endorsed": 100},  # (1,2)->(1,2)
            # B) cluster splintered
            {"shown": 101, "endorsed": 200},  # (3,4,5)->(3,4)
            {"shown": 101, "endorsed": 5},  # (3,4,5)->(5)
            # C) net positive opinion
            {"shown": 102, "endorsed": 102},  # (6,7)->(6,7)
            {"shown": 102, "endorsed": 102},  # repeated...
            {"shown": 102, "endorsed": 6},  # ...and negated: (6,7)->(6)
            {"shown": 102, "endorsed": 7},  # (6,7)->(7)
            # D) net negative opinion
            {"shown": 103, "endorsed": 103},  # (8,9)->(8,9)
            {"shown": 103, "endorsed": 8},  # negated: (8,9)->(8)
            {"shown": 103, "endorsed": 9},  # (8,9)->(9)
            {"shown": 103, "endorsed": 8},  # negated again: (8,9)->(8)
            {"shown": 103, "endorsed": 9},  # (8,9)->(9)
            # E) neutral opinion
            {"shown": 104, "endorsed": 104},  # (10,11)->(10,11)
            {"shown": 104, "endorsed": 10},  # negated: (10,11)->(10)
            {"shown": 104, "endorsed": 11},  # (10,11)->(11)
            # F) judgement has seen less than a previous one
            {"shown": 105, "endorsed": 105},  # (12,13) -> (12,13)
            {"shown": 106, "endorsed": 106},  # (12,13,14) -> (12,13,14)
        ]
    )

    pairs, net_counts, leaves = process_judgements(
        judgements, judgement_cluster_expansion
    )
    assert pairs == {
        (1, 2),
        (3, 4),
        (3, 5),
        (4, 5),
        (6, 7),
        (8, 9),
        (10, 11),
        (12, 13),
        (12, 14),
        (13, 14),
    }

    for pair in [(1, 2), (3, 4), (6, 7), (12, 14), (13, 14)]:
        assert net_counts[pair] == 1

    for pair in [(3, 5), (4, 5), (8, 9)]:
        assert net_counts[pair] == -1

    for pair in [(10, 11)]:
        assert net_counts[pair] == 0

    for pair in [(12, 13)]:
        assert net_counts[pair] == 2

    assert leaves == set(range(1, 15))



"""Unit tests for collection and run management endpoints."""

from typing import Any
from unittest.mock import Mock

import pytest
from fastapi.testclient import TestClient

from matchbox.common.dtos import (
    BackendResourceType,
    Collection,
    Run,
)
from matchbox.common.exceptions import (
    MatchboxCollectionAlreadyExists,
    MatchboxCollectionNotFoundError,
    MatchboxDeletionNotConfirmed,
    MatchboxRunNotFoundError,
)
from matchbox.common.factories.sources import source_factory

# Collection management tests


def test_list_collections(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test listing all collections."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.list_collections = Mock(return_value=["default", "test_collection"])

    response = test_client.get("/collections")

    assert response.status_code == 200
    assert response.json() == ["default", "test_collection"]
    mock_backend.list_collections.assert_called_once()


def test_get_collection(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test retrieving a specific collection."""
    test_client, mock_backend, _ = api_client_and_mocks

    # Mock collection data with runs and resolutions
    collection = Collection(default_run=1, runs=[1, 2])

    mock_backend.get_collection = Mock(return_value=collection)

    response = test_client.get("/collections/test_collection")

    assert response.status_code == 200
    assert Collection.model_validate(response.json())
    mock_backend.get_collection.assert_called_once_with(name="test_collection")


@pytest.mark.parametrize(
    ["method", "endpoint", "backend_method", "extra_params"],
    [
        pytest.param(
            "GET", "/collections/nonexistent", "get_collection", {}, id="get_collection"
        ),
        pytest.param(
            "DELETE",
            "/collections/nonexistent",
            "delete_collection",
            {"certain": True},
            id="delete_collection",
        ),
    ],
)
def test_collection_not_found_404(
    method: str,
    endpoint: str,
    backend_method: str,
    extra_params: dict[str, bool],
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test 404 responses when collection doesn't exist."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_method = getattr(mock_backend, backend_method)
    mock_method.side_effect = MatchboxCollectionNotFoundError()

    response = test_client.request(method, endpoint, params=extra_params)

    assert response.status_code == 404
    assert response.json()["entity"] == BackendResourceType.COLLECTION


def test_create_collection(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test creating a new collection."""
    test_client, mock_backend, _ = api_client_and_mocks

    mock_backend.create_collection = Mock()

    response = test_client.post("/collections/new_collection")

    assert response.status_code == 201
    assert response.json()["success"] is True
    mock_backend.create_collection.assert_called_once_with(name="new_collection")


def test_create_collection_already_exists(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test creating a collection that already exists."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.create_collection = Mock(side_effect=MatchboxCollectionAlreadyExists())

    response = test_client.post("/collections/existing_collection")

    assert response.status_code == 409
    assert response.json()["success"] is False
    assert response.json()["operation"] == "create"
    assert response.json()["details"] == "Collection already exists."


def test_delete_collection(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test deleting a collection with confirmation."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_collection = Mock()

    response = test_client.delete(
        "/collections/test_collection", params={"certain": True}
    )

    assert response.status_code == 200
    assert response.json()["success"] is True
    assert response.json()["operation"] == "delete"
    mock_backend.delete_collection.assert_called_once_with(
        name="test_collection", certain=True
    )


def test_delete_collection_needs_confirmation(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test deleting a collection that requires confirmation."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_collection = Mock(
        side_effect=MatchboxDeletionNotConfirmed(children=[1, 2])
    )

    response = test_client.delete("/collections/test_collection")

    assert response.status_code == 409
    assert response.json()["success"] is False
    assert response.json()["operation"] == "delete"
    assert str(1) in response.json()["details"]
    assert str(2) in response.json()["details"]


# Run management tests


def test_get_run(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test retrieving a specific run."""
    test_client, mock_backend, _ = api_client_and_mocks

    source = source_factory().fake_run().source
    run = Run(
        run_id=1,
        resolutions={source.name: source.to_resolution()},
        is_default=True,
        is_mutable=False,
    )

    mock_backend.get_run = Mock(return_value=run)

    response = test_client.get("/collections/test_collection/runs/1")

    assert response.status_code == 200
    assert Run.model_validate(response.json())
    mock_backend.get_run.assert_called_once_with(collection="test_collection", run_id=1)


@pytest.mark.parametrize(
    [
        "exception_type",
        "expected_entity",
        "method",
        "endpoint",
        "backend_method",
        "extra_params",
    ],
    [
        pytest.param(
            MatchboxCollectionNotFoundError,
            BackendResourceType.COLLECTION,
            "GET",
            "/collections/nonexistent/runs/1",
            "get_run",
            {},
            id="get_run_collection_not_found",
        ),
        pytest.param(
            MatchboxRunNotFoundError,
            BackendResourceType.RUN,
            "GET",
            "/collections/test_collection/runs/13",
            "get_run",
            {},
            id="get_run_run_not_found",
        ),
        pytest.param(
            MatchboxCollectionNotFoundError,
            BackendResourceType.COLLECTION,
            "DELETE",
            "/collections/nonexistent/runs/1",
            "delete_run",
            {"certain": True},
            id="delete_run_collection_not_found",
        ),
        pytest.param(
            MatchboxRunNotFoundError,
            BackendResourceType.RUN,
            "DELETE",
            "/collections/test_collection/runs/13",
            "delete_run",
            {"certain": True},
            id="delete_run_run_not_found",
        ),
    ],
)
def test_run_endpoints_404(
    exception_type: type[Exception],
    expected_entity: BackendResourceType,
    method: str,
    endpoint: str,
    backend_method: str,
    extra_params: dict[str, Any],
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test 404 responses for run endpoints when resources don't exist."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_method = getattr(mock_backend, backend_method)
    mock_method.side_effect = exception_type()

    json_data = {"run_id": 1} if method == "POST" else None
    response = test_client.request(
        method, endpoint, params=extra_params, json=json_data
    )

    assert response.status_code == 404
    assert response.json()["entity"] == expected_entity


def test_create_run(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test creating a new run."""
    test_client, mock_backend, _ = api_client_and_mocks

    source = source_factory().fake_run().source
    run = Run(
        run_id=1,
        resolutions={source.name: source.to_resolution()},
        is_default=False,
        is_mutable=False,
    )

    mock_backend.create_run = Mock(return_value=run)

    response = test_client.post("/collections/test_collection/runs")

    assert response.status_code == 201
    assert Run.model_validate(response.json())
    mock_backend.create_run.assert_called_once_with(collection="test_collection")


def test_set_run_mutable(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test setting run mutability."""
    test_client, mock_backend, _ = api_client_and_mocks

    updated_run = Run(run_id=1, is_default=False, is_mutable=False, resolutions={})
    mock_backend.set_run_mutable = Mock(return_value=updated_run)

    response = test_client.patch(
        "/collections/test_collection/runs/1/mutable", json=False
    )

    assert response.status_code == 200
    assert response.json()["success"] is True
    assert response.json()["operation"] == "update"
    mock_backend.set_run_mutable.assert_called_once_with(
        collection="test_collection", run_id=1, mutable=False
    )


def test_set_run_mutable_missing_flag(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test setting run mutability without providing the flag."""
    test_client, _, _ = api_client_and_mocks

    response = test_client.patch("/collections/test_collection/runs/1/mutable", json={})

    assert response.status_code == 422


def test_set_run_default(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test setting run as default."""
    test_client, mock_backend, _ = api_client_and_mocks

    updated_run = Run(run_id=1, is_default=True, is_mutable=False, resolutions={})
    mock_backend.set_run_default = Mock(return_value=updated_run)

    response = test_client.patch(
        "/collections/test_collection/runs/1/default", json=True
    )

    assert response.status_code == 200
    assert response.json()["success"] is True
    assert response.json()["operation"] == "update"
    mock_backend.set_run_default.assert_called_once_with(
        collection="test_collection", run_id=1, default=True
    )


def test_set_run_default_missing_flag(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test setting run default without providing the flag."""
    test_client, _, _ = api_client_and_mocks

    response = test_client.patch("/collections/test_collection/runs/1/default", json={})

    assert response.status_code == 422


@pytest.mark.parametrize(
    ["endpoint", "payload", "exception_type", "expected_entity"],
    [
        pytest.param(
            "mutable",
            True,
            MatchboxCollectionNotFoundError,
            BackendResourceType.COLLECTION,
        ),
        pytest.param(
            "mutable", True, MatchboxRunNotFoundError, BackendResourceType.RUN
        ),
        pytest.param(
            "default",
            False,
            MatchboxCollectionNotFoundError,
            BackendResourceType.COLLECTION,
        ),
        pytest.param(
            "default", False, MatchboxRunNotFoundError, BackendResourceType.RUN
        ),
    ],
)
def test_run_patch_endpoints_404(
    endpoint: str,
    payload: bool,
    exception_type: type[Exception],
    expected_entity: BackendResourceType,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test 404 responses for run PATCH endpoints when resource doesn't exist."""
    test_client, mock_backend, _ = api_client_and_mocks
    method_name = f"set_run_{endpoint}"
    mock_method = getattr(mock_backend, method_name)
    mock_method.side_effect = exception_type()

    response = test_client.patch(
        f"/collections/test_collection/runs/1/{endpoint}", json=payload
    )

    assert response.status_code == 404
    assert response.json()["entity"] == expected_entity


def test_delete_run(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test deleting a run with confirmation."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_run = Mock()

    response = test_client.delete(
        "/collections/test_collection/runs/1", params={"certain": True}
    )

    assert response.status_code == 200
    assert response.json()["success"] is True
    assert response.json()["operation"] == "delete"
    mock_backend.delete_run.assert_called_once_with(
        collection="test_collection", run_id=1, certain=True
    )


def test_delete_run_needs_confirmation(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test deleting a run that requires confirmation."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_run = Mock(
        side_effect=MatchboxDeletionNotConfirmed(children=[1, 2])
    )

    response = test_client.delete("/collections/test_collection/runs/1")

    assert response.status_code == 409
    assert response.json()["success"] is False
    assert response.json()["operation"] == "delete"
    assert str(1) in response.json()["details"]
    assert str(2) in response.json()["details"]



from collections.abc import Callable, Generator
from os import environ
from unittest.mock import Mock

import pytest
import respx
from fastapi.testclient import TestClient
from httpx import Client
from pydantic import SecretStr
from respx import MockRouter

from matchbox.client._handler import create_client
from matchbox.client._settings import ClientSettings
from matchbox.client._settings import settings as client_settings
from matchbox.client.authorisation import (
    generate_EdDSA_key_pair,
    generate_json_web_token,
)
from matchbox.server.api import app, dependencies
from matchbox.server.base import MatchboxBackends, MatchboxServerSettings
from matchbox.server.uploads import InMemoryUploadTracker


@pytest.fixture(scope="function")
def env_setter() -> Generator[Callable[[str, str], None], None, None]:
    """Set temporary env variable and refresh client settings."""
    original_values = {}

    def setter(var_name: str, var_value: str) -> None:
        original_values[var_name] = environ.get(var_name)

        environ[var_name] = var_value
        client_settings.__init__()

    yield setter

    for var_name, original_value in original_values.items():
        if original_value is None:
            del environ[var_name]
        else:
            environ[var_name] = original_value
    client_settings.__init__()


@pytest.fixture(scope="function")
def api_client_and_mocks(
    env_setter: Callable[[str, str], None],
) -> Generator[tuple[TestClient, Mock, Mock], None, None]:
    """Return client to testable API and associated mocks."""
    # 1) Prepare keys for authentication
    private_key, public_key = generate_EdDSA_key_pair()
    env_setter("MB__CLIENT__PRIVATE_KEY", private_key.decode())
    token = generate_json_web_token(sub="test.user@email.com")
    auth_headers = {"Authorization": token}

    # 2) Override backend with mock
    # Backend has no functionality and must be adapted for each test
    mock_backend = Mock()
    app.dependency_overrides[dependencies.backend] = lambda: mock_backend

    # 3) Override upload tracker with fully functioning mock
    # Note that we don't need to patch the tracker used by the delayed task,
    # as later we set the API as the task runner, and we assume that in that setting
    # the tracker is passed to the background task via dependency injection
    tracker = InMemoryUploadTracker()
    mock_tracker = Mock()
    mock_tracker.get.side_effect = tracker.get
    mock_tracker.set.side_effect = tracker.set
    app.dependency_overrides[dependencies.upload_tracker] = lambda: mock_tracker

    # 4) Override server settings used by API
    test_settings = MatchboxServerSettings(
        backend_type=MatchboxBackends.POSTGRES,
        task_runner="api",
        authorisation=True,
        public_key=SecretStr(public_key.decode()),
    )
    app.dependency_overrides[dependencies.settings] = lambda: test_settings

    # 5) Yield authenticated test client and the server-side mocks
    yield TestClient(app, headers=auth_headers), mock_backend, mock_tracker

    # 6) Restore API's dependencies
    # The client settings should be reset by env_setter once we exit this context
    app.dependency_overrides = {}


@pytest.fixture(scope="function")
def matchbox_api() -> Generator[MockRouter, None, None]:
    """Client for the mocked Matchbox API."""
    with respx.mock(
        base_url=client_settings.api_root, assert_all_called=True
    ) as respx_mock:
        yield respx_mock


@pytest.fixture(scope="session")
def matchbox_client_settings() -> ClientSettings:
    """Client settings for the Matchbox API running in Docker."""
    return client_settings


@pytest.fixture(scope="session")
def matchbox_client(matchbox_client_settings: ClientSettings) -> Client:
    """Client for the Matchbox API running in Docker."""
    return create_client(settings=matchbox_client_settings)
